import astropy.units as u
import numpy as np
import pytest
from astropy.coordinates import SkyCoord
from astropy.cosmology import FlatLambdaCDM
from lightcurvelynx.astro_utils.snia_utils import (
    DistModFromRedshift,
    HostmassX1Distr,
    HostmassX1Func,
    SNCoordGivenPhysicalSep,
)
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from scipy.stats import norm


def test_hostmass_x1_distr():
    """Test that we can create and evaluate a hostmass to X1 distribution."""
    x1_vals = np.arange(-6.0, 6.0, 0.5)

    # At hostmass < 10.0, the distribution should be truncated at -5.0 and 5.0.
    expected = np.array(
        [
            0.0,
            0.0,
            0.0,
            2.5657549651628195e-14,
            1.5060607839169267e-11,
            4.175889141873483e-09,
            5.469335442591314e-07,
            3.383758018885159e-05,
            0.000988879045909292,
            0.01365105415030207,
            0.08901605491595148,
            0.2741887521763265,
            0.3989422804014327,
            0.3520653267642995,
            0.24197072451914337,
            0.12951759566589174,
            0.05399096651318806,
            0.01752830049356854,
            0.0044318484119380075,
            0.0008726826950457602,
            0.00013383022576488537,
            1.5983741106905478e-05,
            0.0,
            0.0,
        ]
    )
    for hostmass in np.arange(5.0, 9.0):
        dist = HostmassX1Distr(hostmass=hostmass)
        pdf_vals = dist.pdf(x1_vals)
        assert np.allclose(pdf_vals, expected)

    # At hostmass >= 10.0 the result should just be a normal with mean 0.0 and scale=1.0.
    expected2 = norm.pdf(x1_vals, loc=0, scale=1)
    for hostmass in np.arange(10.0, 12.0):
        dist = HostmassX1Distr(hostmass=hostmass)
        pdf_vals2 = dist.pdf(x1_vals)
        assert np.allclose(pdf_vals2, expected2)


def test_sample_hostmass_x1c():
    """Test that we can sample correctly from HostmassX1Func."""
    num_samples = 5

    hm_node1 = HostmassX1Func(
        hostmass=NumpyRandomFunc("uniform", low=7, high=12, seed=100),
        seed=101,
    )
    states1 = hm_node1.sample_parameters(num_samples=num_samples)
    values1 = hm_node1.get_param(states1, "function_node_result")
    assert len(values1) == num_samples
    assert len(np.unique(values1)) == num_samples

    # If we are only querying a single sample, we get a float.
    states_single = hm_node1.sample_parameters(num_samples=1)
    values_single = hm_node1.get_param(states_single, "function_node_result")
    assert isinstance(values_single, float)

    # If we create a new node with the same hostmas and the same seeds, we get the
    # same results and the same hostmasses.
    hm_node2 = HostmassX1Func(
        hostmass=NumpyRandomFunc("uniform", low=7, high=12, seed=100),
        seed=101,
    )
    states2 = hm_node2.sample_parameters(num_samples=num_samples)
    values2 = hm_node2.get_param(states2, "function_node_result")
    assert np.allclose(values1, values2)
    assert np.allclose(
        hm_node1.get_param(states1, "hostmass"),
        hm_node2.get_param(states2, "hostmass"),
    )

    # If we use a different seed for the hostmass function only, we get
    # different results (i.e. the hostmass parameter is being resampled).
    hm_node3 = HostmassX1Func(
        hostmass=NumpyRandomFunc("uniform", low=7, high=12, seed=102),
        seed=101,
    )
    states3 = hm_node3.sample_parameters(num_samples=num_samples)
    values3 = hm_node3.get_param(states3, "function_node_result")
    assert not np.allclose(values1, values3)
    assert not np.allclose(
        hm_node1.get_param(states1, "hostmass"),
        hm_node3.get_param(states3, "hostmass"),
    )


def test_dist_mod_from_redshift():
    """Test the computation of dist_mod from the redshift."""
    redshifts = [0.01, 0.02, 0.05, 0.5]
    expected = [33.08419428, 34.60580484, 36.64346629, 42.17006132]

    for idx, z in enumerate(redshifts):
        node = DistModFromRedshift(redshift=z, H0=73.0, Omega_m=0.3)
        state = node.sample_parameters(num_samples=1)
        assert node.get_param(state, "function_node_result") == pytest.approx(expected[idx])

    # Both H0 and Omega_m must be constants.
    rand_node = NumpyRandomFunc("uniform", low=0.0, high=1.0)
    with pytest.raises(ValueError):
        DistModFromRedshift(redshift=0.3, H0=rand_node, Omega_m=0.3)
    with pytest.raises(ValueError):
        DistModFromRedshift(redshift=0.3, H0=73.0, Omega_m=rand_node)


def test_sn_coord_given_physical_separation():
    """Test correct sn coor is returned."""

    host_ra = 0.0
    host_dec = 0.0
    physical_sep_kpc = 1.0
    redshift = 0.5
    H0 = 70.0
    Omega_m = 0.3

    sncoord = SNCoordGivenPhysicalSep(
        host_ra, host_dec, physical_sep_kpc, redshift, H0=H0, Omega_m=Omega_m, node_label="sncoord"
    )
    state = sncoord.sample_parameters()
    sn_ra = state["sncoord.ra"][0]
    sn_dec = state["sncoord.dec"][0]
    # calculate physical separation from host coord, sn coor, and redshift
    host = SkyCoord(host_ra * u.deg, host_dec * u.deg)
    sn = SkyCoord(sn_ra * u.deg, sn_dec * u.deg)

    sep = host.separation(sn)

    cosmo = FlatLambdaCDM(H0=H0, Om0=Omega_m)

    DA = cosmo.angular_diameter_distance(redshift).to(u.kpc).value

    calculated_physical_sep = sep.to(u.rad).value * DA

    assert np.isclose(calculated_physical_sep, physical_sep_kpc)



import numpy as np
import pytest
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc


def test_numpy_random_uniform():
    """Test that we can generate numbers from a uniform distribution."""
    np_node = NumpyRandomFunc("uniform", seed=100)

    values = np.array([np_node.generate() for _ in range(10_000)])
    assert len(np.unique(values)) > 10
    assert np.all(values <= 1.0)
    assert np.all(values >= 0.0)
    assert np.abs(np.mean(values) - 0.5) < 0.01

    # If we reuse the seed, we get the same numbers.
    np_node2 = NumpyRandomFunc("uniform", seed=100)
    values2 = np.array([np_node2.generate() for _ in range(10_000)])
    assert np.allclose(values, values2)

    # If we use a different seed, we get different numbers
    np_node2 = NumpyRandomFunc("uniform", seed=101)
    values2 = np.array([np_node2.generate() for _ in range(10_000)])
    assert not np.allclose(values, values2)

    # But we can override the seed and get the same results again.
    np_node2.set_seed(100)
    values2 = np.array([np_node2.generate() for _ in range(10_000)])
    assert np.allclose(values, values2)

    # We can change the range.
    np_node3 = NumpyRandomFunc("uniform", low=10.0, high=20.0, seed=100)
    values = np.array([np_node3.generate() for _ in range(10_000)])
    assert len(np.unique(values)) > 10
    assert np.all(values <= 20.0)
    assert np.all(values >= 10.0)
    assert np.abs(np.mean(values) - 15.0) < 0.5


def test_numpy_random_uniform_multi_samples():
    """Test that we can generate many numbers at once from a uniform distribution."""
    np_node = NumpyRandomFunc("uniform", seed=100)
    state = np_node.sample_parameters(num_samples=10_000)
    samples = np_node.get_param(state, "function_node_result")
    assert len(samples) == 10_000
    assert len(np.unique(samples)) > 1_000
    assert np.abs(np.mean(samples) - 0.5) < 0.01


def test_numpy_random_uniform_multi_dim():
    """Test that we can generate multi-dimensional vectors from a uniform distribution."""
    # Sample size 2 arrays
    np_node = NumpyRandomFunc("uniform", seed=100, size=2)
    state = np_node.sample_parameters(num_samples=10)
    samples = np_node.get_param(state, "function_node_result")
    assert samples.shape == (10, 2)
    assert len(np.unique(samples.flatten())) == 20

    # Sample size (2, 3) arrays.
    np_node = NumpyRandomFunc("uniform", seed=100, size=(2, 3))
    state = np_node.sample_parameters(num_samples=10)
    samples = np_node.get_param(state, "function_node_result")
    assert samples.shape == (10, 2, 3)
    assert len(np.unique(samples.flatten())) == 60

    # If we do not specify a size and use a single sample, we get a float.
    np_node = NumpyRandomFunc("uniform", seed=100)
    state = np_node.sample_parameters(num_samples=1)
    assert np.isscalar(np_node.get_param(state, "function_node_result"))


def test_numpy_random_normal():
    """Test that we can generate numbers from a normal distribution."""
    np_node = NumpyRandomFunc("normal", loc=100.0, scale=10.0, seed=100, node_label="normal1")

    values = np.array([np_node.generate() for _ in range(10_000)])
    assert np.abs(np.mean(values) - 100.0) < 0.5
    assert np.abs(np.std(values) - 10.0) < 0.5

    # If we reuse the seed, we get the same number.
    np_node2 = NumpyRandomFunc("normal", loc=100.0, scale=10.0, seed=100)
    values2 = np.array([np_node2.generate() for _ in range(10_000)])
    assert np.allclose(values, values2)

    # Check that we can get a dependency graph that correctly includes this node.
    dep_graph = np_node.build_dependency_graph()
    assert dep_graph.all_nodes == {"normal1"}
    assert dep_graph.incoming["normal1.function_node_result"] == set(["normal1.loc", "normal1.scale"])
    assert dep_graph.outgoing["normal1.loc"] == set(["normal1.function_node_result"])
    assert dep_graph.outgoing["normal1.scale"] == set(["normal1.function_node_result"])


def test_numpy_random_given_rng():
    """Test that we can generate numbers from a uniform distribution."""
    np_node1 = NumpyRandomFunc("uniform", seed=100, node_label="node1")
    np_node2 = NumpyRandomFunc("uniform", seed=100, node_label="node2")

    # The first value generated is the same because they are using the node's seed.
    assert np_node1.generate() == pytest.approx(np_node2.generate())

    # But if we use a given random number generators, we get different values.
    value1 = np_node1.generate(rng_info=np.random.default_rng(1))
    value2 = np_node2.generate(rng_info=np.random.default_rng(2))
    assert value1 != pytest.approx(value2)


def test_numpy_choice_fails():
    """Test that we cannot use NumpyRandomFunc with a choice distribution."""
    with pytest.raises(ValueError):
        NumpyRandomFunc("choice", a=5)



"""Utility functions for AGN models.

Adapted from https://github.com/RickKessler/SNANA/blob/master/src/gensed_AGN.py
with the authors' permission.
"""

import numpy as np
from citation_compass import cite_function


@cite_function
def eddington_ratio_dist_fun(edd_ratio, galaxy_type="Blue", rng=None, num_samples=1):
    """Sample from the Eddington Ratio Distribution Function for a given galaxy type.

    Based on notebook from: https://github.com/burke86/imbh_forecast/blob/master/var.ipynb
    and the paper: https://ui.adsabs.harvard.edu/abs/2019ApJ...883..139S/abstract
    with parameters selected from: https://iopscience.iop.org/article/10.3847/1538-4357/aa803b/pdf

    References
    ----------
    * Approach: Sartori et. al. 2019 - https://ui.adsabs.harvard.edu/abs/2019ApJ...883..139S/abstract
    * Parameters: Weigel et. al. 2017 - https://iopscience.iop.org/article/10.3847/1538-4357/aa803b/pdf

    Parameters
    ----------
    edd_ratio : float
        The Eddington ratio.
    galaxy_type : str, optional
        The type of galaxy, either 'Red' or 'Blue'.
        Default: 'Blue'
    rng : np.random.Generator, optional
        The random number generator.
        Default: None
    num_samples : int, optional
        The number of samples to draw. If 1 returns a float otherwise returns an array.
        Default: 1

    Returns
    -------
    result : float or np.array
        The Eddington ratio distribution.
    """
    if rng is None:
        rng = np.random.default_rng()

    if galaxy_type.lower() == "red":
        xi = 10**-2.13
        lambda_br = 10 ** rng.normal(-2.81, 0.18, num_samples)
        delta1 = rng.normal(0.41 - 0.7, 0.02, num_samples)  # > -0.45 won't affect LF
        delta2 = rng.normal(1.22, 0.16, num_samples)
    elif galaxy_type.lower() == "blue":
        xi = 10**-1.65
        lambda_br = 10 ** rng.normal(-1.84, 0.335, num_samples)
        delta1 = rng.normal(0.471 - 0.7, 0.02, num_samples)  # > -0.45 won't affect LF
        delta2 = rng.normal(2.53, 0.53, num_samples)
    else:
        raise ValueError("galaxy_type must be either 'Red' or 'Blue'.")

    result = xi * ((edd_ratio / lambda_br) ** delta1 + (edd_ratio / lambda_br) ** delta2)

    if num_samples == 1:
        return result[0]
    return result



import numpy as np
import pandas as pd
import pytest
from lightcurvelynx.astro_utils.mag_flux import flux2mag
from lightcurvelynx.utils.post_process_results import (
    augment_single_lightcurve,
    concat_results,
    lightcurve_compute_mag,
    lightcurve_compute_snr,
    results_augment_lightcurves,
    results_drop_empty,
)
from nested_pandas import NestedFrame


def test_concat_results():
    """Test the concat_results function."""
    # Create two NestedFrames to concatenate.
    outer_dict_1 = {
        "id": [0, 1, 2],
        "ra": [10.0, 20.0, 30.0],
        "dec": [-10.0, 0.0, 10.0],
        "nobs": [3, 2, 1],
        "z": [0.1, 0.2, 0.3],
    }
    inner_dict_1 = {
        "mjd": [59000, 59001, 59002, 59000, 59001, 59000],
        "flux": [10.0, 12.0, 11.0, 15.0, 14.0, 13.0],
        "fluxerr": [1.0, 1.0, 1.0, 1.5, 1.5, 1.0],
        "filter": ["g", "r", "i", "g", "r", "i"],
    }
    nested_inds_1 = [0, 0, 0, 1, 1, 2]
    res1 = NestedFrame(data=outer_dict_1, index=[0, 1, 2])
    nested_1 = pd.DataFrame(data=inner_dict_1, index=nested_inds_1)
    res1 = res1.add_nested(nested_1, "lightcurve")

    outer_dict_2 = {
        "id": [3, 4],
        "ra": [40.0, 50.0],
        "dec": [-40.0, -50.0],
        "nobs": [1, 2],
        "z": [0.4, 0.5],
    }
    inner_dict_2 = {
        "mjd": [59000, 59001, 59001],
        "flux": [20.0, 22.0, 21.0],
        "fluxerr": [2.0, 2.0, 2.0],
        "filter": ["g", "r", "i"],
    }
    nested_inds_2 = [0, 1, 1]
    res2 = NestedFrame(data=outer_dict_2, index=[0, 1])
    nested_2 = pd.DataFrame(data=inner_dict_2, index=nested_inds_2)
    res2 = res2.add_nested(nested_2, "lightcurve")

    # Concatenate the results.
    results = concat_results([res1, res2])
    assert len(results) == 5
    assert results.columns.tolist() == ["id", "ra", "dec", "nobs", "z", "lightcurve"]

    # Check the outer columns.
    assert results["id"].tolist() == [0, 1, 2, 3, 4]
    assert np.array_equal(results["ra"], [10.0, 20.0, 30.0, 40.0, 50.0])
    assert np.array_equal(results["dec"], [-10.0, 0.0, 10.0, -40.0, -50.0])
    assert np.array_equal(results["z"], [0.1, 0.2, 0.3, 0.4, 0.5])

    # Check one of the inner columns.
    assert results["lightcurve"].nest.fields == ["mjd", "flux", "fluxerr", "filter"]
    assert np.allclose(results["lightcurve"][0]["flux"], [10.0, 12.0, 11.0])
    assert np.allclose(results["lightcurve"][1]["flux"], [15.0, 14.0])
    assert np.allclose(results["lightcurve"][2]["flux"], [13.0])
    assert np.allclose(results["lightcurve"][3]["flux"], [20.0])
    assert np.allclose(results["lightcurve"][4]["flux"], [22.0, 21.0])


def _allclose(a, b, rtol=1e-05, atol=1e-08):
    """Helper function to compare two arrays, treating NaNs and Nones as equal.

    Parameters
    ----------
    a : array-like
        First array to compare.
    b : array-like
        Second array to compare.
    rtol : float, optional
        Relative tolerance. Default is 1e-5.
    atol : float, optional
        Absolute tolerance. Default is 1e-8.
    """
    a = np.ma.asarray(a)
    b = np.ma.asarray(b)

    # Check whether the same entries are None.
    a_valid = (a is not None) & (a is not np.ma.masked)
    b_valid = (b is not None) & (b is not np.ma.masked)
    if not np.array_equal(a_valid, b_valid):
        return False

    a_float = a[a_valid].astype(float)
    b_float = b[b_valid].astype(float)
    return np.allclose(a_float, b_float, rtol=rtol, atol=atol, equal_nan=True)


def test_results_drop_empty():
    """Test the results_drop_empty function."""
    # Create a NestedFrame with some empty lightcurves.
    source_data = {
        "object_id": [0, 1, 2, 3, 4],
        "ra": [10.0, 20.0, 30.0, 40.0, 50.0],
        "dec": [-10.0, -20.0, -30.0, -40.0, -50.0],
        "nobs": [3, 0, 1, 2, 0],
        "z": [0.1, 0.2, 0.3, 0.4, 0.5],
    }
    results = NestedFrame(data=source_data, index=[0, 1, 2, 3, 4])

    # Create a nested DataFrame with lightcurves, some of which are empty.
    nested_data = {
        "mjd": [59000, 59001, 59002, 59003, 59004, 59005],
        "flux": [10.0, 12.0, 11.0, 20.0, 22.0, 30.0],
        "fluxerr": [1.0, 1.0, 1.0, 2.0, 2.0, 3.0],
        "filter": ["g", "r", "i", "g", "r", "i"],
    }
    nested_frame = pd.DataFrame(data=nested_data, index=[0, 0, 0, 2, 3, 3])

    # Add the nested DataFrame to the results.
    results = results.add_nested(nested_frame, "lightcurve")
    assert len(results) == 5
    assert results["object_id"].tolist() == [0, 1, 2, 3, 4]

    # Apply the drop_empty function.
    filtered_results = results_drop_empty(results)
    assert len(filtered_results) == 3
    assert filtered_results["object_id"].tolist() == [0, 2, 3]

    assert len(filtered_results["lightcurve"][0]) == 3
    assert len(filtered_results["lightcurve"][2]) == 1
    assert len(filtered_results["lightcurve"][3]) == 2


def test_lightcurve_compute_snr():
    """Test the lightcurve_compute_snr function."""
    flux = [10.0, 12.0, 0.1, 0.2, 5.0, 6.0, -1.0, 1.0, 10.0]
    fluxerr = [1.0, 1.0, 1.0, 20.0, 2.0, 1.0, 1.0, -1.0, 0.0]
    snr = lightcurve_compute_snr(flux, fluxerr)
    assert _allclose(snr, [10.0, 12.0, 0.1, 0.01, 2.5, 6.0, None, None, None])


def test_lightcurve_compute_mag():
    """Test the lightcurve_compute_mag function."""
    flux = [10.0, 12.0, 0.1, 0.2, 5.0, 6.0, -1.0, 0.0]
    fluxerr = [1.0, 1.0, 1.0, 20.0, 2.0, 1.0, 1.0, 10.0]
    mag, magerr = lightcurve_compute_mag(flux, fluxerr)
    assert _allclose(mag, [flux2mag(f) for f in flux[:-2]] + [None, None])
    assert _allclose(
        magerr,
        [(2.5 / np.log(10)) * (f_err / f_val) for f_val, f_err in zip(flux[:-2], fluxerr[:-2], strict=False)]
        + [None, None],
    )


def test_results_augment_lightcurves():
    """Test the results_augment_lightcurves function."""
    # Create a NestedFrame with some empty lightcurves.
    source_data = {
        "object_id": [0, 1, 2],
        "ra": [10.0, 20.0, 30.0],
        "dec": [-10.0, -20.0, -30.0],
        "nobs": [3, 0, 1],
        "z": [0.1, 0.2, 0.3],
    }
    results = NestedFrame(data=source_data, index=[0, 1, 2])

    with pytest.raises(ValueError):
        # Ensure that the function raises an error if 'lightcurve' is not present.
        results_augment_lightcurves(results, min_snr=5)

    # Create a nested DataFrame with lightcurves, some of which are empty.
    nested_data = {
        "mjd": [59000, 59001, 59002, 59003, 59004, 59005],
        "flux": [10.0, 12.0, 0.1, 0.2, 5.0, 6.0],
        "fluxerr": [1.0, 1.0, 1.0, 20.0, 2.0, 1.0],
        "filter": ["g", "r", "i", "g", "r", "i"],
    }
    nested_frame = pd.DataFrame(data=nested_data, index=[0, 0, 1, 1, 2, 2])

    # Add the nested DataFrame to the results.
    results = results.add_nested(nested_frame, "lightcurve")
    assert len(results) == 3
    assert "snr" not in results["lightcurve"].nest.fields
    assert "detection" not in results["lightcurve"].nest.fields
    assert "mag" not in results["lightcurve"].nest.fields
    assert "magerr" not in results["lightcurve"].nest.fields
    assert results["object_id"].tolist() == [0, 1, 2]

    # Augmenting the lightcurves should add the new columns.
    results_augment_lightcurves(results, min_snr=5)
    assert len(results) == 3

    # Check the SNR and detection markings.
    assert "snr" in results["lightcurve"].nest.fields
    assert _allclose(results["lightcurve.snr"][0].values, [10.0, 12.0])
    assert _allclose(results["lightcurve.snr"][1].values, [0.1, 0.01])
    assert _allclose(results["lightcurve.snr"][2].values, [2.5, 6.0])

    assert "detection" in results["lightcurve"].nest.fields
    assert results["lightcurve.detection"][0].tolist() == [True, True]
    assert results["lightcurve.detection"][1].tolist() == [False, False]
    assert results["lightcurve.detection"][2].tolist() == [False, True]

    # Check the AB magnitudes and magnitude errors.
    assert "mag" in results["lightcurve"].nest.fields
    assert _allclose(results["lightcurve.mag"][0].values, [flux2mag(10.0), flux2mag(12.0)])
    assert _allclose(results["lightcurve.mag"][1].values, [flux2mag(0.1), flux2mag(0.2)])
    assert _allclose(results["lightcurve.mag"][2].values, [flux2mag(5.0), flux2mag(6.0)])

    assert "magerr" in results["lightcurve"].nest.fields
    for i in range(3):
        assert _allclose(
            results["lightcurve.magerr"][i].values,
            (2.5 / np.log(10)) * 1.0 / results["lightcurve.snr"][i].values,
        )

    # Without providing a t0, we do not compute relative time.
    assert "time_rel" not in results["lightcurve"].nest.fields

    # If we provide an array with None, we do not compute relative time.
    results["t0"] = np.array([None, None, None])
    results_augment_lightcurves(results, min_snr=5)
    assert "time_rel" not in results["lightcurve"].nest.fields

    # Try with a valid array of t0.
    results["t0"] = np.array([59000, 59001, 59002])
    results_augment_lightcurves(results, min_snr=5)
    assert "time_rel" in results["lightcurve"].nest.fields
    assert _allclose(results["lightcurve.time_rel"][0].values, [0, 1])
    assert _allclose(results["lightcurve.time_rel"][1].values, [1, 2])
    assert _allclose(results["lightcurve.time_rel"][2].values, [2, 3])


def test_results_augment_lightcurves_single():
    """Test the results_augment_lightcurves function with a non-nested frame."""
    # Create a DataFrame with a lightcurve.
    source_data = {
        "mjd": [59000, 59001, 59002, 59003, 59004, 59005, 59006],
        "flux": [10.0, 12.0, 0.1, 0.2, 5.0, 6.0, -1.0],
    }
    results = pd.DataFrame(data=source_data)

    with pytest.raises(ValueError):
        # Ensure that the function raises an error if 'flux' or 'fluxerr' is not present.
        results_augment_lightcurves(results, min_snr=5)

    # Create a nested DataFrame with lightcurves, some of which are empty.
    results["fluxerr"] = [1.0, 1.0, 1.0, 20.0, 2.0, 1.0, 1.0]

    # Add the nested DataFrame to the results.
    assert "snr" not in results.columns
    assert "detection" not in results.columns
    assert "mag" not in results.columns
    assert "magerr" not in results.columns
    assert "time_rel" not in results.columns

    # Augmenting the lightcurves should add the new columns.
    augment_single_lightcurve(results, min_snr=5)

    # Check we have added the columns.
    assert "snr" in results.columns
    assert "detection" in results.columns
    assert "mag" in results.columns
    assert "magerr" in results.columns
    assert "time_rel" not in results.columns
    assert _allclose(results["snr"].values, [10.0, 12.0, 0.1, 0.01, 2.5, 6.0, None])
    assert np.array_equal(results["detection"].values, [True, True, False, False, False, True, False])
    assert _allclose(
        results["mag"].values,
        [flux2mag(10.0), flux2mag(12.0), flux2mag(0.1), flux2mag(0.2), flux2mag(5.0), flux2mag(6.0), None],
    )
    assert _allclose(results["magerr"].values[:6], (2.5 / np.log(10)) * (1.0 / results["snr"].values[:6]))

    # Try with a t0.
    augment_single_lightcurve(results, min_snr=5, t0=59000)
    assert "time_rel" in results.columns
    assert _allclose(results["time_rel"].values, [0, 1, 2, 3, 4, 5, 6])



"""Additional tests for extrapolation with a time and wavelength bounded model."""

import numpy as np
import pytest
from lightcurvelynx.models.physical_model import SEDModel
from lightcurvelynx.utils.extrapolate import LinearDecay, ZeroPadding


class _LinearLinearTestModel(SEDModel):
    """A model that emits flux as a linear function of wavelength and time:
    f(t, w) = 0.5 * w + 2.0 * t + 100.0

    The model includes bounds on valid times and wavelengths to test extrapolation.

    Parameters
    ----------
    min_phase : float
        The minimum phase of the model (time relative to t0).
    max_phase : float
        The maximum phase of the model (time relative to t0).
    min_wave : float
        The minimum wavelength of the model (in angstroms).
    max_wave : float
        The maximum wavelength of the model (in angstroms).
    fail_on_out_of_bounds : bool
        Whether to raise an error if times or wavelengths are out of bounds.
    **kwargs : dict
        Additional keyword arguments.
    """

    def __init__(
        self,
        min_phase=0.0,
        max_phase=100.0,
        min_wave=1000.0,
        max_wave=12000.0,
        fail_on_out_of_bounds=True,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.min_phase = min_phase
        self.max_phase = max_phase
        self.min_wave = min_wave
        self.max_wave = max_wave
        self.fail_on_out_of_bounds = fail_on_out_of_bounds

    def minwave(self, **kwargs):
        """Get the minimum wavelength of the model."""
        return self.min_wave

    def maxwave(self, **kwargs):
        """Get the maximum wavelength of the model."""
        return self.max_wave

    def minphase(self, **kwargs):
        """Get the minimum phase of the model."""
        return self.min_phase

    def maxphase(self, **kwargs):
        """Get the maximum phase of the model."""
        return self.max_phase

    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        t0 = self.get_param(graph_state, "t0")
        if t0 is None:
            raise ValueError("t0 parameter is required for this model.")

        phase = times - t0
        if self.fail_on_out_of_bounds and (np.any(phase < self.min_phase) or np.any(phase > self.max_phase)):
            raise ValueError("Times are out of bounds for this model.")
        if self.fail_on_out_of_bounds and (
            np.any(wavelengths < self.min_wave) or np.any(wavelengths > self.max_wave)
        ):
            raise ValueError("Wavelengths are out of bounds for this model.")

        time_component = 2.0 * phase[:, np.newaxis]
        wavelength_component = 0.5 * wavelengths[np.newaxis, :]
        return 100.0 + time_component + wavelength_component


def test_linear_linear_model() -> None:
    """Test the _LinearLinearTestModel with valid times and wavelengths."""
    query_waves = np.array([1500.0, 2000.0])
    query_times = np.array([0.0, 10.0, 100.0])
    expected = np.array(
        [
            [100.0 + 2.0 * 0.0 + 0.5 * 1500.0, 100.0 + 2.0 * 0.0 + 0.5 * 2000.0],
            [100.0 + 2.0 * 10.0 + 0.5 * 1500.0, 100.0 + 2.0 * 10.0 + 0.5 * 2000.0],
            [100.0 + 2.0 * 100.0 + 0.5 * 1500.0, 100.0 + 2.0 * 100.0 + 0.5 * 2000.0],
        ]
    )

    model = _LinearLinearTestModel(t0=0.0)
    values = model.evaluate_sed(query_times, query_waves)
    assert values.shape == (3, 2)
    assert np.allclose(values, expected)


def test_linear_linear_model_bounds() -> None:
    """Test the _LinearLinearTestModel with out-of-bounds times and wavelengths."""
    model = _LinearLinearTestModel(t0=0.0)

    # We fail a call evaluate_sed since the times are out of bounds. But we
    # should get a warning about extrapolation first.
    query_times = np.array([-10.0, 50.0, 110.0])
    query_waves = np.array([1500.0, 2000.0])
    with pytest.warns(UserWarning):
        with pytest.raises(ValueError):
            _ = model.evaluate_sed(query_times, query_waves)

    # We fail a call to evaluate_sed with out-of-bounds wavelengths. But we
    # should get a warning about extrapolation first.
    query_times = np.array([10.0, 20.0])
    query_waves = np.array([0.0, 2000.0, 5000.0, 13000.0])
    with pytest.warns(UserWarning):
        with pytest.raises(ValueError):
            _ = model.evaluate_sed(query_times, query_waves)


def test_linear_linear_model_extrapolators() -> None:
    """Test the _LinearLinearTestModel with given extrapolators on
    out-of-bounds times and wavelengths."""
    # Wavelength falls of faster (per step) than time.
    wave_extrapolator = LinearDecay(decay_width=500.0)  # 500 angstroms to zero
    time_extrapolator = LinearDecay(decay_width=100.0)  # 100 days to zero
    model = _LinearLinearTestModel(
        wave_extrapolation=wave_extrapolator,
        time_extrapolation=time_extrapolator,
        t0=0.0,
    )

    # Three query wavelengths: one before (80% factor) and two inside.
    query_waves = np.array([900.0, 2000.0, 5000.0])
    # Three query times: one before (90% factor), three inside, and one after (80% factor).
    query_times = np.array([-10.0, 0.0, 50.0, 100.0, 120.0])
    values = model.evaluate_sed(query_times, query_waves)
    print(values)

    expected = np.array(
        [
            [432.0, 990.0, 2340.0],
            [480.0, 1100.0, 2600.0],
            [560.0, 1200.0, 2700.0],
            [640.0, 1300.0, 2800.0],
            [512.0, 1040.0, 2240.0],
        ]
    )
    assert np.allclose(values, expected)


def test_linear_linear_model_diff_extrapolators() -> None:
    """Test the _LinearLinearTestModel with different extrapolators each
    each direction (above/below) for times and wavelengths.
    """
    wave_linear = LinearDecay(decay_width=500.0)  # 500 angstroms to zero
    time_linear = LinearDecay(decay_width=100.0)  # 100 days to zero
    zero_extrap = ZeroPadding()

    # Three query wavelengths: one before, two inside, and one after.
    query_waves = np.array([900.0, 2000.0, 5000.0, 12100.0])
    # Three query times: one before, three inside, and one after .
    query_times = np.array([-10.0, 0.0, 50.0, 100.0, 120.0])

    # Start to (zero, linear) for wavelength, and (linear, zero) for time.
    model = _LinearLinearTestModel(
        wave_extrapolation=(zero_extrap, wave_linear),
        time_extrapolation=(time_linear, zero_extrap),
        t0=0.0,
    )
    values = model.evaluate_sed(query_times, query_waves)
    expected = np.array(
        [
            [0.0, 990.0, 2340.0, 4392.0],
            [0.0, 1100.0, 2600.0, 4880.0],
            [0.0, 1200.0, 2700.0, 4960.0],
            [0.0, 1300.0, 2800.0, 5040.0],
            [0.0, 0.0, 0.0, 0.0],
        ]
    )
    assert np.allclose(values, expected)

    # Start to (linear, zero) for wavelength, and (linear, linear) for time.
    model = _LinearLinearTestModel(
        wave_extrapolation=(wave_linear, zero_extrap),
        time_extrapolation=(time_linear, time_linear),
        t0=0.0,
    )
    values = model.evaluate_sed(query_times, query_waves)
    print(values)
    expected = np.array(
        [
            [432.0, 990.0, 2340.0, 0.0],
            [480.0, 1100.0, 2600.0, 0.0],
            [560.0, 1200.0, 2700.0, 0.0],
            [640.0, 1300.0, 2800.0, 0.0],
            [512.0, 1040.0, 2240.0, 0.0],
        ]
    )
    assert np.allclose(values, expected)

    # We fail if we use any Nones since the model doesn't know how to extrapolate.
    model = _LinearLinearTestModel(
        wave_extrapolation=(wave_linear, zero_extrap),
        time_extrapolation=(None, time_linear),
        t0=0.0,
    )
    with pytest.warns(UserWarning):
        with pytest.raises(ValueError):
            _ = model.evaluate_sed(query_times, query_waves)

    model = _LinearLinearTestModel(
        wave_extrapolation=(None, zero_extrap),
        time_extrapolation=(time_linear, time_linear),
        t0=0.0,
    )
    with pytest.warns(UserWarning):
        with pytest.raises(ValueError):
            _ = model.evaluate_sed(query_times, query_waves)


def test_linear_linear_model_ooo_time() -> None:
    """Test the _LinearLinearTestModel with extrapolators on out-of-order times."""
    time_linear = LinearDecay(decay_width=100.0)  # 100 days to zero

    query_waves = np.array([2000.0])
    query_times = np.array([-10.0, 0.0, -15.0, 50.0, 125.0, 100.0, 75.0, 120.0])

    # Start to (zero, linear) for wavelength, and (linear, zero) for time.
    model = _LinearLinearTestModel(
        time_extrapolation=(time_linear, time_linear),
        t0=0.0,
    )
    values = model.evaluate_sed(query_times, query_waves)
    expected = np.array([[990.0], [1100.0], [935.0], [1200.0], [975.0], [1300.0], [1250.0], [1040.0]])
    assert np.allclose(values, expected)

    # With no extrapolation (and no failing), we should just query the model.
    model2 = _LinearLinearTestModel(
        time_extrapolation=(None, None),
        t0=0.0,
        fail_on_out_of_bounds=False,
    )
    with pytest.warns(UserWarning):
        values2 = model2.evaluate_sed(query_times, query_waves)
    expected2 = np.array([[1080.0], [1100.0], [1070.0], [1200.0], [1350.0], [1300.0], [1250.0], [1340.0]])
    assert np.allclose(values2, expected2)


def test_linear_linear_model_ooo_wavelength() -> None:
    """Test the _LinearLinearTestModel with extrapolators on out-of-order wavelengths."""
    wave_linear = LinearDecay(decay_width=500.0)  # 500 angstroms to zero

    query_waves = np.array([900.0, 1000.0, 2000.0, 800.0, 12200.0, 5000.0, 12100.0])
    query_times = np.array([50.0])

    # Start to (zero, linear) for wavelength, and (linear, zero) for time.
    model = _LinearLinearTestModel(
        wave_extrapolation=(wave_linear, wave_linear),
        t0=0.0,
    )
    values = model.evaluate_sed(query_times, query_waves)
    expected = np.array([560.0, 700.0, 1200.0, 420.0, 3720.0, 2700.0, 4960.0])
    assert np.allclose(values, expected)

    # With no extrapolation (and no failing), we should just query the model.
    model2 = _LinearLinearTestModel(
        wave_extrapolation=(None, wave_linear),
        t0=0.0,
        fail_on_out_of_bounds=False,
    )
    with pytest.warns(UserWarning):
        values2 = model2.evaluate_sed(query_times, query_waves)
    expected2 = np.array([650.0, 700.0, 1200.0, 600.0, 3720.0, 2700.0, 4960.0])
    assert np.allclose(values2, expected2)

    # With no extrapolation (and no failing), we should just query the model.
    model3 = _LinearLinearTestModel(
        wave_extrapolation=(wave_linear, None),
        t0=0.0,
        fail_on_out_of_bounds=False,
    )
    with pytest.warns(UserWarning):
        values3 = model3.evaluate_sed(query_times, query_waves)
    expected3 = np.array([560.0, 700.0, 1200.0, 420.0, 6300.0, 2700.0, 6250.0])
    assert np.allclose(values3, expected3)



import numpy as np
import pandas as pd
import pytest
from lightcurvelynx.consts import GAUSS_EFF_AREA2FWHM_SQ
from lightcurvelynx.obstable.fake_obs_table import FakeObsTable


def test_create_fake_obs_table_consts():
    """Create a minimal FakeObsTable object with given defaults."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    pdf = pd.DataFrame(values)

    zp_per_band = {"g": 26.0, "r": 27.0, "i": 28.0}
    ops_data = FakeObsTable(
        pdf,
        noise_strategy="exhaustive",
        zp_per_band=zp_per_band,
        fwhm_px=2.0,
        sky_bg_electrons=100.0,
    )
    assert len(ops_data) == 5

    # We use the defaults when we do not provide values in the table. Not all of these
    # will be added as columns, but we can still retrieve an array of values.
    assert np.allclose(ops_data["ra"], values["ra"])
    assert np.allclose(ops_data["dec"], values["dec"])
    assert np.allclose(ops_data["time"], values["time"])
    assert np.array_equal(ops_data["filter"], values["filter"])
    assert np.allclose(ops_data["zp"], [27.0, 26.0, 27.0, 28.0, 26.0])
    assert np.allclose(ops_data.get_value_per_row("fwhm_px"), [2.0] * 5)
    assert np.allclose(ops_data.get_value_per_row("sky_bg_electrons"), [100.0] * 5)
    assert np.allclose(ops_data.get_value_per_row("exptime"), [30.0] * 5)
    assert np.allclose(ops_data.get_value_per_row("nexposure"), [1] * 5)

    # Derived from fwhm_px.
    assert np.allclose(
        ops_data.get_value_per_row("psf_footprint"),
        [GAUSS_EFF_AREA2FWHM_SQ * (2.0) ** 2] * 5,
    )

    assert ops_data.survey_values["dark_current"] == 0
    assert ops_data.survey_values["nexposure"] == 1
    assert ops_data.survey_values["sky_bg_electrons"] == 100
    assert ops_data.survey_values["radius"] is None
    assert ops_data.survey_values["read_noise"] == 0
    assert ops_data.survey_values["survey_name"] == "FAKE_SURVEY"

    # Successfully derived a zp column.
    assert "zp" in ops_data

    # We can compute noise.
    flux_error = ops_data.bandflux_error_point_source(
        np.array([100.0, 200.0, 300.0, 400.0, 500.0]),  # Fluxes in nJy
        np.arange(5),  # Indices of the observations
    )
    assert len(flux_error) == 5
    assert np.all(flux_error > 0)
    assert len(np.unique(flux_error)) > 1  # Not all the same

    # If we give psf_footprint, we use that instead of fwhm_px.
    ops_data = FakeObsTable(
        pdf,
        noise_strategy="exhaustive",
        zp_per_band=zp_per_band,
        fwhm_px=2.0,
        psf_footprint=1.0,
        sky_bg_electrons=100.0,
    )
    assert np.allclose(ops_data.get_value_per_row("psf_footprint"), [1.0] * 5)

    # We can override the defaults, using dictionaries of values for fwhm_px and sky.
    ops_data = FakeObsTable(
        pdf,
        noise_strategy="exhaustive",
        zp_per_band=zp_per_band,
        exptime=60.0,
        fwhm_px={"g": 2.5, "r": 3.1, "i": 1.9},
        nexposure=2,
        radius=1.0,
        read_noise=5.0,
        sky_bg_electrons={"g": 150.0, "r": 140.0, "i": 155.0},
        survey_name="MY_SURVEY",
    )
    assert len(ops_data) == 5

    # We use the defaults when we do not provide values in the table. Not all of these
    # will be added as columns, but we can still retrieve an array of values.
    assert np.allclose(ops_data["ra"], values["ra"])
    assert np.allclose(ops_data["dec"], values["dec"])
    assert np.allclose(ops_data["time"], values["time"])
    assert np.array_equal(ops_data["filter"], values["filter"])
    assert np.allclose(ops_data["zp"], [27.0, 26.0, 27.0, 28.0, 26.0])
    assert np.allclose(ops_data.get_value_per_row("fwhm_px"), [3.1, 2.5, 3.1, 1.9, 2.5])
    assert np.allclose(ops_data.get_value_per_row("sky_bg_electrons"), [140.0, 150.0, 140.0, 155.0, 150.0])
    assert np.allclose(ops_data.get_value_per_row("exptime"), [60.0] * 5)
    assert np.allclose(ops_data.get_value_per_row("nexposure"), [2] * 5)
    assert np.allclose(
        ops_data.get_value_per_row("psf_footprint"),
        [
            GAUSS_EFF_AREA2FWHM_SQ * (3.1) ** 2,
            GAUSS_EFF_AREA2FWHM_SQ * (2.5) ** 2,
            GAUSS_EFF_AREA2FWHM_SQ * (3.1) ** 2,
            GAUSS_EFF_AREA2FWHM_SQ * (1.9) ** 2,
            GAUSS_EFF_AREA2FWHM_SQ * (2.5) ** 2,
        ],
    )

    assert ops_data.survey_values["dark_current"] == 0
    assert ops_data.survey_values["nexposure"] == 2
    assert ops_data.survey_values["sky_bg_electrons"] == {"g": 150.0, "r": 140.0, "i": 155.0}
    assert ops_data.survey_values["radius"] == 1.0
    assert ops_data.survey_values["read_noise"] == 5.0
    assert ops_data.survey_values["survey_name"] == "MY_SURVEY"

    # We can compute noise.
    flux_error2 = ops_data.bandflux_error_point_source(
        np.array([100.0, 200.0, 300.0, 400.0, 500.0]),  # Fluxes in nJy
        np.arange(5),  # Indices of the observations
    )
    assert len(flux_error2) == 5
    assert np.all(flux_error2 > 0)
    assert len(np.unique(flux_error2)) > 1  # Not all the same
    assert np.any(flux_error2 != flux_error)  # Different from before

    # We fail if we use a deriver that cannot compute required parameters.
    with pytest.raises(ValueError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="given_only",
            zp_per_band=zp_per_band,
            fwhm_px=2.0,
            psf_footprint=1.0,
            sky_bg_electrons=100.0,
        )


def test_create_fake_obs_table_non_consts():
    """Test that if we specify values in columns, we use those instead of the defaults."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
        "fwhm_px": np.array([2.0, 2.5, 3.0, 2.0, 1.5]),
        "sky_bg_electrons": np.array([100.0, 150.0, 200.0, 100.0, 50.0]),
        "exptime": np.array([20.0, 30.0, 40.0, 20.0, 10.0]),
        "nexposure": np.array([1, 2, 3, 1, 1]),
        "zp": np.array([10.0, 11.0, 12.0, 13.0, 14.0]),
    }
    pdf = pd.DataFrame(values)

    zp_per_band = {"g": 26.0, "r": 27.0, "i": 28.0}
    ops_data = FakeObsTable(
        pdf,
        noise_strategy="exhaustive",
        zp_per_band=zp_per_band,
        fwhm_px=2.0,
        sky=100.0,
    )
    assert len(ops_data) == 5
    assert np.allclose(ops_data["ra"], values["ra"])
    assert np.allclose(ops_data["dec"], values["dec"])
    assert np.allclose(ops_data["time"], values["time"])
    assert np.array_equal(ops_data["filter"], values["filter"])
    assert np.allclose(ops_data["fwhm_px"], values["fwhm_px"])
    assert np.allclose(ops_data["sky_bg_electrons"], values["sky_bg_electrons"])
    assert np.allclose(ops_data["exptime"], values["exptime"])
    assert np.allclose(ops_data["nexposure"], values["nexposure"])
    assert np.allclose(ops_data["zp"], values["zp"])

    # We can compute noise.
    flux_error = ops_data.bandflux_error_point_source(
        np.array([100.0, 200.0, 300.0, 400.0, 500.0]),  # Fluxes in nJy
        np.arange(5),  # Indices of the observations
    )
    assert len(flux_error) == 5
    assert np.all(flux_error > 0)
    assert len(np.unique(flux_error)) > 1  # Not all the same


def test_create_fake_obs_table_cols_fail():
    """Test that we raise errors when we do not provide required values."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    pdf = pd.DataFrame(values)

    # Missing fwhm_px.
    zp_per_band = None
    with pytest.raises(ValueError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="exhaustive",
            zp_per_band=zp_per_band,
            sky_bg_electrons=100.0,
        )

    # Missing sky_bg_electrons.
    with pytest.raises(ValueError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="exhaustive",
            zp_per_band=zp_per_band,
            fwhm_px=2.0,
        )

    # Missing or invalid exptime.
    with pytest.raises(ValueError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="exhaustive",
            zp_per_band=zp_per_band,
            fwhm_px=2.0,
            sky_bg_electrons=100.0,
            exptime=None,
        )
    with pytest.raises(ValueError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="exhaustive",
            zp_per_band=zp_per_band,
            fwhm_px=2.0,
            sky_bg_electrons=100.0,
            exptime=-10.0,
        )

    # Missing or invalid nexposure.
    with pytest.raises(ValueError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="exhaustive",
            zp_per_band=zp_per_band,
            fwhm_px=2.0,
            sky_bg_electrons=100.0,
            nexposure=None,
        )
    with pytest.raises(ValueError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="exhaustive",
            zp_per_band=zp_per_band,
            fwhm_px=2.0,
            sky_bg_electrons=100.0,
            nexposure=-1,
        )


def test_create_fake_obs_table_zp_fail():
    """Test that we raise errors when we do not provide required values."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
    }
    pdf = pd.DataFrame(values)

    # No filters from which to computer zp.
    zp_per_band = {"g": 26.0, "r": 27.0, "i": 28.0}
    with pytest.raises(KeyError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="exhaustive",
            zp_per_band=zp_per_band,
            fwhm_px=2.0,
            sky=100.0,
        )

    # Mismatched filters.
    values["filter"] = np.array(["r", "g", "r", "i", "z"])
    pdf = pd.DataFrame(values)

    zp_per_band = {"g": 26.0, "r": 27.0}
    with pytest.raises(ValueError):
        _ = FakeObsTable(
            pdf,
            noise_strategy="exhaustive",
            zp_per_band=zp_per_band,
            fwhm_px=2.0,
            sky=100.0,
        )


def test_create_fake_obs_table_noise_free():
    """Create a noise free FakeObsTable object."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    pdf = pd.DataFrame(values)

    ops_data = FakeObsTable(pdf, const_flux_error=0.0)
    assert len(ops_data) == 5

    # Compute the bandflux error for each observation (should all be zero).
    fluxes = np.array([100.0, 200.0, 300.0, 400.0, 500.0])  # Fluxes in nJy
    flux_error = ops_data.bandflux_error_point_source(fluxes, np.arange(5))
    assert np.array_equal(flux_error, np.zeros(5))



"""The top-level module for survey related data, such as pointing and noise
information. By default the module uses the Rubin OpSim data, but it can be
extended to other survey data as well.
"""

from __future__ import annotations  # "type1 | type2" syntax in Python <3.10

import numpy as np
import pandas as pd

from lightcurvelynx import _LIGHTCURVELYNX_BASE_DATA_DIR
from lightcurvelynx.astro_utils.mag_flux import mag2flux
from lightcurvelynx.astro_utils.noise_model import poisson_bandflux_std
from lightcurvelynx.astro_utils.zeropoint import flux_electron_zeropoint
from lightcurvelynx.consts import GAUSS_EFF_AREA2FWHM_SQ
from lightcurvelynx.obstable.obs_table import ObsTable
from lightcurvelynx.utils.data_download import download_data_file_if_needed

LSSTCAM_PIXEL_SCALE = 0.2
"""The pixel scale for the LSST camera in arcseconds per pixel."""

_lsstcam_readout_noise = 8.8
"""The standard deviation of the count of readout electrons per pixel for the LSST camera.

The value is from https://smtn-002.lsst.io/v/OPSIM-1171/index.html
"""

_lsstcam_dark_current = 0.2
"""The dark current for the LSST camera in electrons per second per pixel.

The value is from https://smtn-002.lsst.io/v/OPSIM-1171/index.html
"""

_lsstcam_view_radius = 1.75
"""The angular radius of the observation field (in degrees)."""

_lsst_zp_err_mag = 1.0e-4
"""The zero point error in magnitude.

We choose a very conservative noise flooring of 1e-4 mag.
This number will be updated when we have a better estimate from LSST.
"""

_lsstcam_extinction_coeff = {
    "u": -0.458,
    "g": -0.208,
    "r": -0.122,
    "i": -0.074,
    "z": -0.057,
    "y": -0.095,
}
"""The extinction coefficients for the LSST filters.

Values are from
https://community.lsst.org/t/release-of-v3-4-simulations/8548/12
Calculated with syseng_throughputs v1.9
"""

_lsstcam_zeropoint_per_sec_zenith = {
    "u": 26.524,
    "g": 28.508,
    "r": 28.361,
    "i": 28.171,
    "z": 27.782,
    "y": 26.818,
}
"""The zeropoints for the LSST filters at zenith

This is magnitude that produces 1 electron in a 1 second exposure,
see _assign_zero_points() docs for more details.

Values are from
https://community.lsst.org/t/release-of-v3-4-simulations/8548/12
Calculated with syseng_throughputs v1.9
"""


class OpSim(ObsTable):
    """A wrapper class around the opsim table with cached data for efficiency.

    Parameters
    ----------
    table : dict or pandas.core.frame.DataFrame
        The table with all the OpSim information.
    colmap : dict
        A mapping of short column names to their names in the underlying table.
        Defaults to the Rubin OpSim column names, stored in the class variable
        _opsim_colnames.
    saturation_mags : dict, optional
        A dictionary mapping filter names to their saturation thresholds in magnitudes. The filters
        provided must match those in the table. If not provided, OpSim-specific defaults will be
        used.
    **kwargs : dict
        Additional keyword arguments to pass to the constructor. This includes overrides
        for survey parameters such as:
        - dark_current : The dark current for the camera in electrons per second per pixel.
        - ext_coeff: Mapping of filter names to extinction coefficients.
        - pixel_scale: The pixel scale for the camera in arcseconds per pixel.
        - radius: The angular radius of the observations (in degrees).
        - read_noise: The readout noise for the camera in electrons per pixel.
        - zp_per_sec: Mapping of filter names to zeropoints at zenith.
    """

    _required_names = ["ra", "dec", "time"]

    # Default column names for the Rubin OpSim.
    _default_colnames = {
        "airmass": "airmass",
        "dec": "fieldDec",
        "exptime": "visitExposureTime",
        "filter": "filter",
        "ra": "fieldRA",
        "time": "observationStartMJD",
        "zp": "zp_nJy",  # We add this column to the table
        "seeing": "seeingFwhmEff",
        "skybrightness": "skyBrightness",
        "nexposure": "numExposures",
    }

    # Default survey values.
    _default_survey_values = {
        "dark_current": _lsstcam_dark_current,
        "ext_coeff": _lsstcam_extinction_coeff,
        "pixel_scale": LSSTCAM_PIXEL_SCALE,
        "radius": _lsstcam_view_radius,
        "read_noise": _lsstcam_readout_noise,
        "zp_per_sec": _lsstcam_zeropoint_per_sec_zenith,
        "zp_err_mag": _lsst_zp_err_mag,
        "survey_name": "LSST",
    }

    # Default LSST saturation thresholds in magnitudes.
    # https://www.lsst.org/sites/default/files/docs/sciencebook/SB_3.pdf
    _default_saturation_mags = {
        "u": 14.7,
        "g": 15.7,
        "r": 15.8,
        "i": 15.8,
        "z": 15.3,
        "y": 13.9,
    }

    # Class constants for the column names.
    def __init__(
        self,
        table,
        colmap=None,
        saturation_mags=None,
        **kwargs,
    ):
        colmap = self._default_colnames if colmap is None else colmap

        # If saturation thresholds are not provided, then set to the OpSim defaults.
        if saturation_mags is None:
            saturation_mags = self._default_saturation_mags

        super().__init__(table, colmap=colmap, saturation_mags=saturation_mags, **kwargs)

    def _assign_zero_points(self):
        """Assign instrumental zero points in nJy to the OpSim tables."""
        cols = self._table.columns.to_list()
        if not ("filter" in cols and "airmass" in cols and "exptime" in cols):
            raise ValueError(
                "OpSim does not include the columns needed to derive zero point "
                "information. Required columns: filter, airmass, and exptime."
            )

        zp_values = flux_electron_zeropoint(
            ext_coeff=self.safe_get_survey_value("ext_coeff"),
            instr_zp_mag=self.safe_get_survey_value("zp_per_sec"),
            filter=self._table["filter"],
            airmass=self._table["airmass"],
            exptime=self._table["exptime"],
        )
        self.add_column("zp", zp_values, overwrite=True)

    @classmethod
    def from_url(cls, opsim_url, force_download=False):
        """Construct an OpSim object from a URL to a predefined opsim data file.

        For Rubin OpSim data, you will typically use the latest baseline data set in:
        https://s3df.slac.stanford.edu/data/rubin/sim-data/
        such as:
        https://s3df.slac.stanford.edu/data/rubin/sim-data/sims_featureScheduler_runs3.4/baseline/baseline_v3.4_10yrs.db

        Parameters
        ----------
        opsim_url : str
            The URL to the opsim data file.
        force_download : bool, optional
            If True, the OpSim data will be downloaded even if it already exists locally.
            Default is False.

        Returns
        -------
        opsim : OpSim
            An OpSim object containing the data from the specified URL.
        """
        data_file_name = opsim_url.split("/")[-1]
        data_path = _LIGHTCURVELYNX_BASE_DATA_DIR / "opsim" / data_file_name

        if not download_data_file_if_needed(data_path, opsim_url, force_download=force_download):
            raise RuntimeError(f"Failed to download opsim data from {opsim_url}.")
        return cls.from_db(data_path)

    def bandflux_error_point_source(self, bandflux, index):
        """Compute observational bandflux error for a point source

        Parameters
        ----------
        bandflux : array_like of float
            Band bandflux of the point source in nJy.
        index : array_like of int
            The index of the observation in the OpSim table.

        Returns
        -------
        flux_err : array_like of float
            Simulated bandflux noise in nJy.
        """
        observations = self._table.iloc[index]

        # By the effective FWHM definition, see
        # https://smtn-002.lsst.io/v/OPSIM-1171/index.html
        # We need it in pixel^2
        pixel_scale = self.safe_get_survey_value("pixel_scale")
        psf_footprint = GAUSS_EFF_AREA2FWHM_SQ * (observations["seeing"] / pixel_scale) ** 2
        zp = observations["zp"]

        # Table value is in mag/arcsec^2
        sky_njy_angular = mag2flux(observations["skybrightness"])
        # We need electrons per pixel^2
        sky = sky_njy_angular * pixel_scale**2 / zp

        return poisson_bandflux_std(
            bandflux,
            total_exposure_time=observations["exptime"],
            exposure_count=observations["nexposure"],
            psf_footprint=psf_footprint,
            sky=sky,
            zp=zp,
            readout_noise=self.safe_get_survey_value("read_noise"),
            dark_current=self.safe_get_survey_value("dark_current"),
            zp_err_mag=self.safe_get_survey_value("zp_err_mag"),
        )


def create_random_opsim(num_obs, seed=None):
    """Create a random OpSim pointings drawn uniformly from (RA, dec).

    Parameters
    ----------
    num_obs : int
        The size of the OpSim to generate.
    seed : int
        The seed to used for random number generation. If None then
        uses a default random number generator.
        Default: None

    Returns
    -------
    opsim_data : OpSim
        The OpSim data structure.
    seed : int, optional
        The seed for the random number generator.
    """
    if num_obs <= 0:
        raise ValueError("Number of observations must be greater than zero.")

    rng = np.random.default_rng() if seed is None else np.random.default_rng(seed=seed)

    # Generate the (RA, dec) pairs uniformly on the surface of a sphere.
    ra = np.degrees(rng.uniform(0.0, 2.0 * np.pi, size=num_obs))
    dec = np.degrees(np.arccos(2.0 * rng.uniform(0.0, 1.0, size=num_obs) - 1.0) - (np.pi / 2.0))

    # Generate the information needed to compute zeropoint.
    airmass = rng.uniform(1.3, 1.7, size=num_obs)
    filter = rng.choice(["u", "g", "r", "i", "z", "y"], size=num_obs)

    input_data = {
        "observationStartMJD": 0.05 * np.arange(num_obs),
        "fieldRA": ra,
        "fieldDec": dec,
        "airmass": airmass,
        "filter": filter,
        "visitExposureTime": 29.0 * np.ones(num_obs),
    }

    opsim = OpSim(input_data)
    return opsim


def opsim_add_random_data(opsim_data, colname, min_val=0.0, max_val=1.0):
    """Add a column composed of random uniform data. Used for testing.

    Parameters
    ----------
    opsim_data : OpSim
        The OpSim data structure to modify.
    colname : str
        The name of the new column to add.
    min_val : float
        The minimum value of the uniform range.
        Default: 0.0
    max_val : float
        The maximum value of the uniform range.
        Default: 1.0
    """
    values = np.random.uniform(low=min_val, high=max_val, size=len(opsim_data))
    opsim_data.add_column(colname, values)


def oversample_opsim(
    opsim: OpSim,
    *,
    pointing: tuple[float, float] = (200, -50),
    search_radius: float = 1.75,
    delta_t: float = 0.01,
    time_range: tuple[float | None, float | None] = (None, None),
    bands: list[str] | None = None,
    strategy: str = "darkest_sky",
):
    """Single-pointing oversampled OpSim table.

    It includes observations for a single pointing only,
    but with very high time resolution. The observations
    would alternate between the bands.

    Parameters
    ----------
    opsim : OpSim
        The OpSim table to oversample.
    pointing : tuple of RA and Dec in degrees
        The pointing to use for the oversampled table.
    search_radius : float, optional
        The search radius for the oversampled table in degrees.
        The default is the half of the LSST's field of view.
    delta_t : float, optional
        The time between observations in days.
    time_range : tuple or floats or Nones, optional
        The start and end times of the observations in MJD.
        None means to use the minimum (maximum) time in
        all the observations found for the given pointing.
        Time is being samples as np.arange(*time_range, delta_t).
    bands : list of str or None, optional
        The list of bands to include in the oversampled table.
        The default is to include all bands found for the given pointing.
    strategy : str, optional
        The strategy to select prototype observations.
        - "darkest_sky" selects the observations with the minimal sky brightness
          (maximum "skyBrightness" value) in each band. This is the default.
        - "random" selects the observations randomly. Fixed seed is used.

    """
    ra, dec = pointing
    observations = opsim._table.iloc[opsim.range_search(ra, dec, radius=search_radius)]
    if len(observations) == 0:
        raise ValueError("No observations found for the given pointing.")

    time_min, time_max = time_range
    if time_min is None:
        time_min = np.min(observations["time"])
    if time_max is None:
        time_max = np.max(observations["time"])
    if time_min >= time_max:
        raise ValueError(f"Invalid time_range: start > end: {time_min} > {time_max}")

    uniq_bands = np.unique(observations["filter"])
    if bands is None:
        bands = uniq_bands
    elif not set(bands).issubset(uniq_bands):
        raise ValueError(f"Invalid bands: {bands}")

    new_times = np.arange(time_min, time_max, delta_t)
    n = len(new_times)
    if n < len(bands):
        raise ValueError("Not enough time points to cover all bands.")

    new_table = pd.DataFrame(
        {
            # Just in case, to not have confusion with the original table
            "observationId": opsim._table["observationId"].max() + 1 + np.arange(n),
            "time": new_times,
            "ra": ra,
            "dec": dec,
            "filter": np.tile(bands, n // len(bands)),
        }
    )
    other_columns = [column for column in observations.columns if column not in new_table.columns]

    if strategy == "darkest_sky":
        for band in bands:
            # MAXimum magnitude is MINimum brightness (darkest sky)
            idxmax = observations["skybrightness"][observations["filter"] == band].idxmax()
            idx = new_table.index[new_table["filter"] == band]
            darkest_sky_obs = pd.DataFrame.from_records([observations.loc[idxmax]] * idx.size, index=idx)
            new_table.loc[idx, other_columns] = darkest_sky_obs[other_columns]
    elif strategy == "random":
        rng = np.random.default_rng(0)
        for band in bands:
            single_band_obs = observations[observations["filter"] == band]
            idx = new_table.index[new_table["filter"] == band]
            random_obs = single_band_obs.sample(idx.size, replace=True, random_state=rng).set_index(idx)
            new_table.loc[idx, other_columns] = random_obs[other_columns]
    else:
        raise ValueError(f"Invalid strategy: {strategy}")

    return OpSim(
        new_table,
        colmap=opsim._colmap,
        **opsim.survey_values,
    )



"""The base classes for all models.

The code supports two types of models: 1) SEDModels define recipes for computing SEDs
at given times and wavelengths, accounting for redshift and other effects.
2) BandfluxModels only compute band fluxes for specific passbands instead of the SEDs. This is used for models
that are empirically fit from observed band fluxes.

We strongly recommend using the full SED models (SEDModels) whenever possible since they
more accurately simulate aspects such as the impact of redshift on rest frame effects.
"""

import warnings
from abc import ABC
from os import urandom

import numpy as np

from lightcurvelynx.astro_utils.passbands import Passband, PassbandGroup
from lightcurvelynx.astro_utils.redshift import RedshiftDistFunc, obs_to_rest_times_waves, rest_to_obs_flux
from lightcurvelynx.base_models import ParameterizedNode
from lightcurvelynx.utils.extrapolate import FluxExtrapolationModel


class BasePhysicalModel(ParameterizedNode, ABC):
    """The abstract base class used to represent a physical model of a source of flux. This includes
    basic attributes, such as right ascension, declination, redshift, and distance.

    Physical models can have fixed attributes (where you need to create a new model or use
    a setter function to change them) and settable model parameters that can be passed functions
    or constants and are stored in the graph's (external) graph_state dictionary.

    Physical models also support adding and applying a variety of effects, such as redshift.

    Parameterized values include:
      * dec - The object's declination in degrees.
      * distance - The object's luminosity distance in pc.
      * ra - The object's right ascension in degrees.
      * redshift - The object's redshift.
      * t0 - The t0 of the zero phase (if applicable), date.

    Parameters
    ----------
    ra : float
        The object's right ascension (in degrees)
    dec : float
        The object's declination (in degrees)
    redshift : float
        The object's redshift.
    t0 : float
        The phase offset in MJD. For non-time-varying phenomena, this has no effect.
    distance : float
        The object's luminosity distance (in pc). If no value is provided and
        a cosmology parameter is given, the model will try to derive from
        the redshift and the cosmology.
    node_label : str, optional
        The label for the node in the model graph.
    seed : int, optional
        The seed for a random number generator.
    **kwargs : dict, optional
        Any additional keyword arguments.
    """

    def __init__(
        self,
        *,
        ra=None,
        dec=None,
        redshift=None,
        t0=None,
        distance=None,
        node_label=None,
        seed=None,
        **kwargs,
    ):
        super().__init__(node_label=node_label, **kwargs)

        # Set the parameters for the model.
        self.add_parameter(
            "ra", ra, description="The object's right ascension (in degrees)", allow_gradient=False
        )
        self.add_parameter(
            "dec", dec, description="The object's declination (in degrees)", allow_gradient=False
        )
        self.add_parameter("redshift", redshift, description="The object's redshift.", allow_gradient=False)
        self.add_parameter("t0", t0, description="The phase offset in MJD.")

        # If the luminosity distance is provided, use that. Otherwise try the
        # redshift value using the cosmology (if given). Finally, default to None.
        if distance is not None:
            self.add_parameter(
                "distance",
                distance,
                description="The object's luminosity distance (in pc)",
                allow_gradient=False,
            )
        elif redshift is not None and kwargs.get("cosmology") is not None:
            self._redshift_func = RedshiftDistFunc(redshift=self.redshift, **kwargs)
            self.add_parameter(
                "distance",
                self._redshift_func,
                description="The object's luminosity distance (in pc)",
                allow_gradient=False,
            )
        else:
            self.add_parameter(
                "distance", None, description="The object's luminosity distance (in pc)", allow_gradient=False
            )

        # Get a default random number generator for this object, using the
        # given seed if one is provided.
        if seed is None:
            seed = int.from_bytes(urandom(4), "big")
        self._rng = np.random.default_rng(seed=seed)

    def minwave(self, **kwargs):
        """Get the minimum supported wavelength of the model.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        minwave : float or None
            The minimum wavelength of the model (in angstroms) or None
            if the model does not have a defined minimum wavelength.
        """
        return None

    def maxwave(self, **kwargs):
        """Get the maximum supported wavelength of the model.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        maximum : float or None
            The maximum wavelength of the model (in angstroms) or None
            if the model does not have a defined maximum wavelength.
        """
        return None

    def minphase(self, **kwargs):
        """Get the minimum supported phase of the model in days.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        minphase : float or None
            The minimum phase of the model (in days) or None
            if the model does not have a defined minimum phase.
        """
        return None

    def maxphase(self, **kwargs):
        """Get the maximum supported phase of the model in days.

        Parameters
        ----------
        **kwargs : dict
            Additional keyword arguments, not used in this method.

        Returns
        -------
        maximum : float or None
            The maximum phase of the model (in days) or None
            if the model does not have a defined maximum phase.
        """
        return None

    def add_effect(self, effect):
        """Add an effect to the model. This effect will be applied to all
        fluxes densities simulated by the model.

        Any effect parameters that are not already in the model
        will be added to this node's parameters.

        Parameters
        ----------
        effect : EffectModel
            The effect to add.
        """
        raise NotImplementedError

    def mask_by_time(self, times, graph_state=None):
        """Compute a mask for whether a given time is of interest for a given object.
        For example, a user can use this function to generate a mask to include
        only the observations of interest for a window around the supernova.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        graph_state : GraphState, optional
            An object mapping graph parameters to their values.

        Returns
        -------
        time_mask : numpy.ndarray
            A length T array of Booleans indicating whether the time is of interest.
        """
        return np.full(len(times), True)

    def evaluate_bandfluxes(self, passband_or_group, times, filters, state, rng_info=None) -> np.ndarray:
        """Get the band fluxes for a given Passband or PassbandGroup.

        Parameters
        ----------
        passband_or_group : Passband or PassbandGroup
            The passband (or passband group) to use.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray or None
            A length T array of filter names. It may be None if
            passband_or_group is a Passband.
        state : GraphState
            An object mapping graph parameters to their values.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A matrix of the band fluxes. If only one sample is provided in the GraphState,
            then returns a length T array. Otherwise returns a size S x T array where S is the
            number of samples in the graph state.
        """
        # Check if we need to sample the graph.
        if state is None:
            state = self.sample_parameters(num_samples=1, rng_info=rng_info)

        if isinstance(passband_or_group, Passband):
            # If we are just given a passband, turn it into a passband group and save
            # the list of the filter name (repeated).
            passband_group = PassbandGroup([passband_or_group])
            if filters is None:
                filters = np.full(len(times), passband_or_group.filter_name)
        else:
            # This could be a PassbandGroup or (in limited cases) None.
            passband_group = passband_or_group

        if filters is None:
            raise ValueError("If passband_or_group is a PassbandGroup, filters must be provided.")
        filters = np.asarray(filters)
        if len(filters) != len(times):
            raise ValueError("Filters array must have the same length as times array.")

        # If we only have a single sample, we can return the band fluxes directly.
        if state.num_samples == 1:
            return self._evaluate_bandfluxes_single(passband_group, times, filters, state, rng_info=rng_info)

        # Fill in the band fluxes one at a time and return them all.
        bandfluxes = np.empty((state.num_samples, len(times)))
        for sample_num, current_state in enumerate(state):
            current_fluxes = self._evaluate_bandfluxes_single(
                passband_group,
                times,
                filters,
                current_state,
                rng_info=rng_info,
            )
            bandfluxes[sample_num, :] = current_fluxes[np.newaxis, :]
        return bandfluxes


class SEDModel(BasePhysicalModel):
    """A model of a source of flux that is defined at the SED level.

    Attributes
    ----------
    rest_frame_effects : list of EffectModel
        A list of effects to apply in the rest frame.
    obs_frame_effects : list of EffectModel
        A list of effects to apply in the observer frame.
    apply_redshift : bool
        Whether to apply redshift to the model.

    Parameters
    ----------
    wave_extrapolation : FluxExtrapolationModel or tuple, optional
        The extrapolation model(s) to use for wavelengths that fall outside the model's defined
        bounds. If a tuple is provided, then it is expected to be of the form (before_model, after_model)
        where before_model is the model for before the first valid wavelength and after_model is
        the model for after the last valid wavelength. If None is provided the model will not try to
        extrapolate, but rather call compute_sed() for all wavelengths.
    time_extrapolation : FluxExtrapolationModel or tuple, optional
        The extrapolation model(s) to use for times that fall outside the model's defined
        bounds. If a tuple is provided, then it is expected to be of the form (before_model, after_model)
        where before_model is the model for before the first valid time and after_model is
        the model for after the last valid time. If None is provided the model will not try to
        extrapolate, but rather call compute_sed() for all times.
    """

    def __init__(
        self,
        wave_extrapolation=None,
        time_extrapolation=None,
        *args,
        **kwargs,
    ):
        super().__init__(*args, **kwargs)

        # Initialize the effect settings to their default values.
        self.apply_redshift = kwargs.get("redshift") is not None
        self.rest_frame_effects = []
        self.obs_frame_effects = []

        # Set the extrapolation for values outside the model's defined bounds.
        if wave_extrapolation is None:
            self._wave_extrap_before = None
            self._wave_extrap_after = None
        elif isinstance(wave_extrapolation, tuple):
            if len(wave_extrapolation) != 2:
                raise ValueError("If wave_extrapolation is a tuple, it must have length 2.")
            self._wave_extrap_before = wave_extrapolation[0]
            self._wave_extrap_after = wave_extrapolation[1]
        elif isinstance(wave_extrapolation, FluxExtrapolationModel):
            self._wave_extrap_before = wave_extrapolation
            self._wave_extrap_after = wave_extrapolation
        else:
            raise TypeError("wave_extrapolation must be a FluxExtrapolationModel or a tuple of two models.")

        if time_extrapolation is None:
            self._time_extrap_before = None
            self._time_extrap_after = None
        elif isinstance(time_extrapolation, tuple):
            if len(time_extrapolation) != 2:
                raise ValueError("If time_extrapolation is a tuple, it must have length 2.")
            self._time_extrap_before = time_extrapolation[0]
            self._time_extrap_after = time_extrapolation[1]
        elif isinstance(time_extrapolation, FluxExtrapolationModel):
            self._time_extrap_before = time_extrapolation
            self._time_extrap_after = time_extrapolation
        else:
            raise TypeError("time_extrapolation must be a FluxExtrapolationModel or a tuple of two models.")

    def set_apply_redshift(self, apply_redshift):
        """Toggles the apply_redshift setting. If set to True, the model will
        apply redshift during the flux density computation including applying wavelength
        and time transformations.

        Parameters
        ----------
        apply_redshift : bool
            The new value for apply_redshift.
        """
        self.apply_redshift = apply_redshift

    def add_effect(self, effect, skip_params=False):
        """Add an effect to the model. This effect will be applied to all
        fluxes densities simulated by the model.

        Any effect parameters that are not already in the model
        will be added to this node's parameters.

        Parameters
        ----------
        effect : EffectModel
            The effect to add.
        skip_params : bool
            Skip adding the parameters to the model. This should only be done
            in very limited cases where the parameters are added via another mechanism.
            Most users should NOT change this setting.
            Default: False
        """
        # Add any effect parameters that are not already in the model.
        if not skip_params:
            for param_name, setter in effect.parameters.items():
                if param_name not in self.setters:
                    self.add_parameter(
                        param_name,
                        setter,
                        description=f"Added parameter by effect {effect}",
                        allow_gradient=False,
                    )

        # Add the effect to the appropriate list.
        if effect.rest_frame:
            self.rest_frame_effects.append(effect)
        else:
            self.obs_frame_effects.append(effect)

    def list_effects(self):
        """Return a list of all effects in the order in which they are applied."""
        return self.rest_frame_effects + self.obs_frame_effects

    def compute_sed(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free rest frame flux densities.
        The rest-frame flux is defined as F_nu = L_nu / 4*pi*D_L**2,
        where D_L is the luminosity distance.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps in MJD.
        wavelengths : numpy.ndarray, optional
            A length N array of rest frame wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
            Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of rest frame SED values (in nJy).
        """
        raise NotImplementedError()

    def compute_sed_with_extrapolation(self, times, wavelengths, graph_state, **kwargs):
        """Draw effect-free observations for this object, extrapolating
        to times and wavelengths where the model is not defined.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of rest frame timestamps.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState
            An object mapping graph parameters to their values.
        **kwargs : dict, optional
           Any additional keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of SED values (in nJy).
        """
        query_waves = np.copy(wavelengths)
        query_times = np.copy(times)

        # We check if we can do extrapolation for before the first valid wavelength and, if so, modify
        # the queries and set up the data we need.
        min_query_wave = np.min(wavelengths)
        min_valid_wave = self.minwave(graph_state=graph_state)
        if min_valid_wave is None:
            min_valid_wave = min_query_wave

        before_wave_queries = None
        if min_query_wave < min_valid_wave:
            if self._wave_extrap_before is None:
                warnings.warn(
                    "Some wavelengths are less than the model's defined bounds and no wave "
                    "extrapolation is set. If this is not the intended, you can enable wavelength "
                    "extrapolation using the 'wave_extrapolation' parameter."
                )
            else:
                # Add the boundary point at the start for extrapolation and compute
                # the list of wavelengths to extrapolate.
                valid_mask = query_waves >= min_valid_wave
                before_wave_queries = query_waves[~valid_mask]
                query_waves = np.concatenate(([min_valid_wave], query_waves[valid_mask]))

        # We check if we can do extrapolation for after the last valid wavelength and, if so, modify
        # the queries and set up the data we need.
        max_query_wave = np.max(wavelengths)
        max_valid_wave = self.maxwave(graph_state=graph_state)
        if max_valid_wave is None:
            max_valid_wave = max_query_wave

        after_wave_queries = None
        if max_query_wave > max_valid_wave:
            if self._wave_extrap_after is None:
                warnings.warn(
                    "Some wavelengths are greater than the model's defined bounds and no wave "
                    "extrapolation is set. If this is not the intended, you can enable wavelength "
                    "extrapolation using the 'wave_extrapolation' parameter."
                )
            else:
                # Add the boundary point at the end for extrapolation and compute
                # the list of wavelengths to extrapolate.
                valid_mask = query_waves <= max_valid_wave
                after_wave_queries = query_waves[~valid_mask]
                query_waves = np.concatenate((query_waves[valid_mask], [max_valid_wave]))

        # Get t0 offset since the time bounds are given in phase.
        t0 = self.get_param(graph_state, "t0")
        if t0 is None:
            t0 = 0.0

        # We check if we can do extrapolation for times before the valid time range and, if so, modify
        # the queries and set up the data we need.
        min_query_time = np.min(times)
        min_valid_phase = self.minphase(graph_state=graph_state)
        if min_valid_phase is None:
            min_valid_time = min_query_time
        else:
            min_valid_time = min_valid_phase + t0

        before_time_queries = None
        if min_query_time < min_valid_time:
            if self._time_extrap_before is None:
                warnings.warn(
                    "Some times are less than the model's defined bounds and no time "
                    "extrapolation is set. If this is not the intended, you can enable time "
                    "extrapolation using the 'time_extrapolation' parameter."
                )
            else:
                # Add the boundary point at the start for extrapolation and compute
                # the list of times to extrapolate.
                valid_mask = query_times >= min_valid_time
                before_time_queries = query_times[~valid_mask]
                query_times = np.concatenate(([min_valid_time], query_times[valid_mask]))

        # We check if we can do extrapolation for times after the valid time range and, if so, modify
        # the queries and set up the data we need.
        max_query_time = np.max(times)
        max_valid_phase = self.maxphase(graph_state=graph_state)
        if max_valid_phase is None:
            max_valid_time = max_query_time
        else:
            max_valid_time = max_valid_phase + t0

        after_time_queries = None
        if max_query_time > max_valid_time:
            if self._time_extrap_after is None:
                warnings.warn(
                    "Some times are greater than the model's defined bounds and no time "
                    "extrapolation is set. If this is not the intended, you can enable time "
                    "extrapolation using the 'time_extrapolation' parameter."
                )
            else:
                # Add the boundary point at the end for extrapolation and compute
                # the list of times to extrapolate.
                valid_mask = query_times <= max_valid_time
                after_time_queries = query_times[~valid_mask]
                query_times = np.concatenate((query_times[valid_mask], [max_valid_time]))

        # Get the flux density at all times and wavelengths (except those we will extrapolate).
        computed_flux = self.compute_sed(query_times, query_waves, graph_state)

        # We do the extrapolation in two steps: first for wavelengths and then for times.
        # The result is that we combine the extrapolation for both dimensions at the corners.
        if before_wave_queries is not None or after_wave_queries is not None:
            new_computed_flux = np.zeros((len(query_times), len(wavelengths)))
            in_bounds_mask = np.full(len(wavelengths), True)

            if before_wave_queries is not None:
                # Compute the flux values before the model's first valid wavelength and
                # fill the extrapolated values in the correct locations.
                before_wave_mask = wavelengths < min_valid_wave
                extrapolated_values = self._wave_extrap_before.extrapolate_wavelength(
                    min_valid_wave,
                    computed_flux[:, 0],
                    before_wave_queries,
                )
                new_computed_flux[:, before_wave_mask] = extrapolated_values
                in_bounds_mask[before_wave_mask] = False

                # Drop the first column (which was added for extrapolation).
                computed_flux = computed_flux[:, 1:]

            if after_wave_queries is not None:
                # Compute the flux values after the model's last valid wavelength and
                # fill the extrapolated values in the correct locations.
                after_wave_mask = wavelengths > max_valid_wave
                extrapolated_values = self._wave_extrap_after.extrapolate_wavelength(
                    max_valid_wave,
                    computed_flux[:, -1],
                    after_wave_queries,
                )
                new_computed_flux[:, after_wave_mask] = extrapolated_values
                in_bounds_mask[after_wave_mask] = False

                # Drop the last column (which was added for extrapolation).
                computed_flux = computed_flux[:, :-1]

            # Fill in the non-extrapolated values and rename it to computed_flux.
            new_computed_flux[:, in_bounds_mask] = computed_flux
            computed_flux = new_computed_flux

        # Do a similiar process for time extrapolation.
        if before_time_queries is not None or after_time_queries is not None:
            new_computed_flux = np.zeros((len(times), len(wavelengths)))
            in_bounds_mask = np.full(len(times), True)

            if before_time_queries is not None:
                # Compute the flux values before the model's first valid time and
                # fill the extrapolated values in the correct locations.
                before_time_mask = times < min_valid_time
                extrapolated_values = self._time_extrap_before.extrapolate_time(
                    min_valid_time,
                    computed_flux[0, :],
                    before_time_queries,
                )
                new_computed_flux[before_time_mask, :] = extrapolated_values
                in_bounds_mask[before_time_mask] = False

                # Drop the first row (which was added for extrapolation).
                computed_flux = computed_flux[1:, :]

            if after_time_queries is not None:
                # Compute the flux values after the model's last valid time and
                # fill the extrapolated values in the correct locations.
                after_time_mask = times > max_valid_time
                extrapolated_values = self._time_extrap_after.extrapolate_time(
                    max_valid_time,
                    computed_flux[-1, :],
                    after_time_queries,
                )
                new_computed_flux[after_time_mask, :] = extrapolated_values
                in_bounds_mask[after_time_mask] = False

                # Drop the last row (which was added for extrapolation).
                computed_flux = computed_flux[:-1, :]

            # Fill in the non-extrapolated values and rename it to computed_flux.
            new_computed_flux[in_bounds_mask, :] = computed_flux
            computed_flux = new_computed_flux

        return computed_flux

    def _evaluate_single(self, times, wavelengths, state, rng_info=None, **kwargs):
        """Evaluate the model and apply the effects for a single, given graph state.
        This function applies redshift, computes the flux density for the object,
        applies rest frames effects, performs the redshift correction (if needed),
        and applies the observer frame effects.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.
        """
        if state is None or state.num_samples != 1:
            raise ValueError("A GraphState with num_samples=1 required.")
        params = self.get_local_params(state)

        # Pre-effects are adjustments done to times and/or wavelengths, before flux density
        # computation. We skip if redshift is 0.0 since there is nothing to do.
        if self.apply_redshift and params["redshift"] != 0.0:
            if params.get("redshift", None) is None:
                raise ValueError("The 'redshift' parameter is required for redshifted models.")
            if params.get("t0", None) is None:
                raise ValueError("The 't0' parameter is required for redshifted models.")
            rest_times, rest_wavelengths = obs_to_rest_times_waves(
                times, wavelengths, params["redshift"], params["t0"]
            )
        else:
            rest_times = times
            rest_wavelengths = wavelengths

        # Compute the flux density for the object and apply any rest frame effects.
        flux_density = self.compute_sed_with_extrapolation(rest_times, rest_wavelengths, state, **kwargs)
        for effect in self.rest_frame_effects:
            flux_density = effect.apply(
                flux_density,
                times=rest_times,
                wavelengths=rest_wavelengths,
                rng_info=rng_info,
                **params,  # Provide all the node's parameters to the effect.
            )

        # Post-effects are adjustments done to the flux density after computation.
        if self.apply_redshift and params["redshift"] != 0.0:
            # We have alread checked that redshift is not None.
            flux_density = rest_to_obs_flux(flux_density, params["redshift"])

        # Apply observer frame effects.
        for effect in self.obs_frame_effects:
            flux_density = effect.apply(
                flux_density,
                times=times,
                wavelengths=wavelengths,
                rng_info=rng_info,
                **params,  # Provide all the node's parameters to the effect.
            )
        return flux_density

    def evaluate_sed(self, times, wavelengths, graph_state=None, given_args=None, rng_info=None, **kwargs):
        """Draw observations for this object and apply the noise.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        wavelengths : numpy.ndarray
            A length N array of wavelengths (in angstroms).
        graph_state : GraphState, optional
            An object mapping graph parameters to their values.
        given_args : dict, optional
            A dictionary representing the given arguments for this sample run.
            This can be used as the JAX PyTree for differentiation.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            All the other keyword arguments.

        Returns
        -------
        flux_density : numpy.ndarray
            A length S x T x N matrix of SED values (in nJy), where S is the number of samples,
            T is the number of time steps, and N is the number of wavelengths.
            If S=1 then the function returns a T x N matrix.
        """
        # Make sure times and wavelengths are numpy arrays.
        times = np.asarray(times)
        wavelengths = np.asarray(wavelengths)

        # Check if we need to sample the graph.
        if graph_state is None:
            graph_state = self.sample_parameters(
                given_args=given_args, num_samples=1, rng_info=rng_info, **kwargs
            )

        # If we only have a single sample, do not bother to iterate through the states.
        if graph_state.num_samples == 1:
            return self._evaluate_single(
                times,
                wavelengths,
                graph_state,
                rng_info=rng_info,
                **kwargs,
            )

        # Iterate through each graph state computing the flux for each sample.
        results = np.empty((graph_state.num_samples, len(times), len(wavelengths)))
        for sample_num, state in enumerate(graph_state):
            # Compute the flux (handling redshift and applying all effects)
            # then save the result to the array of all results.
            results[sample_num, :, :] = self._evaluate_single(
                times,
                wavelengths,
                state,
                rng_info=rng_info,
                **kwargs,
            )
        return results

    def _evaluate_bandfluxes_single(self, passband_group, times, filters, state, rng_info=None) -> np.ndarray:
        """Get the band fluxes for a given PassbandGroup and a single, given graph state.

        Parameters
        ----------
        passband_group : PassbandGroup
            The passband group to use.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray
            A length T array of filter names.
        state : GraphState
            An object mapping graph parameters to their values.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes for this sample.
        """
        bandfluxes = np.empty(len(times))
        for filter_name in np.unique(filters):
            # Compute the band fluxes for the times at which this filter is used.
            passband = passband_group[filter_name]
            filter_mask = filters == filter_name

            # Compute the spectral fluxes at the same wavelengths used to define the passband.
            # The evaluate function applies all effects (rest and observation frame) for the source
            # as well as handling all the redshift conversions.
            spectral_fluxes = self.evaluate_sed(times[filter_mask], passband.waves, state, rng_info=rng_info)
            bandfluxes[filter_mask] = passband.fluxes_to_bandflux(spectral_fluxes)
        return bandfluxes


class BandfluxModel(BasePhysicalModel, ABC):
    """A model of a source of flux that is only defined by band pass values
    in the observer frame (instead of a full SED).

    Instead of calling `compute_sed()` the model calls `compute_bandflux()` for each
    filter during its computation.

    Note
    ----
    We strongly recommend using the full SED models (SEDModel) whenever possible
    since they more accurately simulate aspects such as the impact of redshift on rest
    frame effects.

    Attributes
    ----------
    band_pass_effects : list of EffectModel
        A list of effects to apply in to the band pass fluxes.

    Parameters
    ----------
    time_extrapolation : FluxExtrapolationModel or tuple, optional
        The extrapolation model(s) to use for times that fall outside the model's defined
        bounds. If a tuple is provided, then it is expected to be of the form (before_model, after_model)
        where before_model is the model for before the first valid time and after_model is
        the model for after the last valid time. If None is provided the model will not try to
        extrapolate, but rather call compute_bandflux() for all times.
    """

    def __init__(self, *args, time_extrapolation=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.band_pass_effects = []

        if time_extrapolation is None:
            self._time_extrap_before = None
            self._time_extrap_after = None
        elif isinstance(time_extrapolation, tuple):
            if len(time_extrapolation) != 2:
                raise ValueError("If time_extrapolation is a tuple, it must have length 2.")
            self._time_extrap_before = time_extrapolation[0]
            self._time_extrap_after = time_extrapolation[1]
        elif isinstance(time_extrapolation, FluxExtrapolationModel):
            self._time_extrap_before = time_extrapolation
            self._time_extrap_after = time_extrapolation
        else:
            raise TypeError("time_extrapolation must be a FluxExtrapolationModel or a tuple of two models.")

        if "wave_extrapolation" in kwargs and kwargs["wave_extrapolation"] is not None:
            warnings.warn("BandfluxModel does not support wave_extrapolation, but value provided.")

    def set_apply_redshift(self, apply_redshift):
        """Toggles the apply_redshift setting.

        Parameters
        ----------
        apply_redshift : bool
            The new value for apply_redshift.
        """
        raise NotImplementedError("BandfluxModel does not support apply_redshift.")

    def add_effect(self, effect, skip_params=False):
        """Add an effect to the model.

        Parameters
        ----------
        effect : EffectModel
            The effect to add.
        skip_params : bool
            Skip adding the parameters to the model. This should only be done
            in very limited cases where the parameters are added via another mechanism.
            Most users should NOT change this setting.
            Default: False
        """
        # Add any effect parameters that are not already in the model.
        if not skip_params:
            for param_name, setter in effect.parameters.items():
                if param_name not in self.setters:
                    self.add_parameter(
                        param_name,
                        setter,
                        description=f"Added parameter by effect {effect}",
                        allow_gradient=False,
                    )

        # Add the effect to the band pass effects list.
        self.band_pass_effects.append(effect)

    def list_effects(self):
        """Return a list of all effects in the order in which they are applied."""
        return self.band_pass_effects

    def compute_bandflux(self, times, filter, state, rng_info=None):
        """Evaluate the model at the passband level for a single, given graph state and filter.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filter : str
            The name of the filter.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandflux : numpy.ndarray
            A length T array of band fluxes for this model in this filter.
        """
        raise NotImplementedError

    def compute_bandflux_with_extrapolation(self, times, filter, state, rng_info=None):
        """Evaluate the model at the passband level for a single, given graph state and filter,
        extrapolating to times where the model is not defined.

        Parameters
        ----------
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filter : str
            The name of the filter.
        state : GraphState
            An object mapping graph parameters to their values with num_samples=1.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandflux : numpy.ndarray
            A length T array of band fluxes for this model in this filter.
        """
        query_times = np.copy(times)

        # Get t0 offset since the time bounds are given in phase.
        t0 = self.get_param(state, "t0")
        if t0 is None:
            t0 = 0.0

        # We check if we can do extrapolation for times before the valid time range and, if so, modify
        # the queries and set up the data we need.
        min_query_time = np.min(times)
        min_valid_phase = self.minphase(filter=filter, graph_state=state)
        if min_valid_phase is None:
            min_valid_time = min_query_time
        else:
            min_valid_time = min_valid_phase + t0

        before_time_queries = None
        if min_query_time < min_valid_time:
            if self._time_extrap_before is None:
                warnings.warn(
                    "Some times are less than the model's defined bounds and no time "
                    "extrapolation is set. If this is not the intended, you can enable time "
                    "extrapolation using the 'time_extrapolation' parameter."
                )
            else:
                # Add the boundary point at the start for extrapolation and compute
                # the list of times to extrapolate.
                valid_mask = query_times >= min_valid_time
                before_time_queries = query_times[~valid_mask]
                query_times = np.concatenate(([min_valid_time], query_times[valid_mask]))

        # We check if we can do extrapolation for times after the valid time range and, if so, modify
        # the queries and set up the data we need.
        max_query_time = np.max(times)
        max_valid_phase = self.maxphase(filter=filter, graph_state=state)
        if max_valid_phase is None:
            max_valid_time = max_query_time
        else:
            max_valid_time = max_valid_phase + t0

        after_time_queries = None
        if max_query_time > max_valid_time:
            if self._time_extrap_after is None:
                warnings.warn(
                    "Some times are greater than the model's defined bounds and no time "
                    "extrapolation is set. If this is not the intended, you can enable time "
                    "extrapolation using the 'time_extrapolation' parameter."
                )
            else:
                # Add the boundary point at the end for extrapolation and compute
                # the list of times to extrapolate.
                valid_mask = query_times <= max_valid_time
                after_time_queries = query_times[~valid_mask]
                query_times = np.concatenate((query_times[valid_mask], [max_valid_time]))

        # Get the band flux at all times (except those we will extrapolate).
        computed_flux = self.compute_bandflux(query_times, filter, state, rng_info=rng_info)

        # Then do extrapolation for times that fell outside the model's bounds. These might
        # not be in order, so we use masks to keep track of where they go.
        if before_time_queries is not None or after_time_queries is not None:
            new_computed_flux = np.zeros(len(times))
            in_bounds_mask = np.full(len(times), True)

            if before_time_queries is not None:
                # Compute the flux values before the model's first valid time.
                before_time_mask = times < min_valid_time
                extrapolated_values = self._time_extrap_before.extrapolate_time(
                    min_valid_time,
                    np.array([computed_flux[0]]),
                    before_time_queries,
                )
                new_computed_flux[before_time_mask] = extrapolated_values[:, 0]
                in_bounds_mask[before_time_mask] = False

                # Drop the first entry (which was added for extrapolation).
                computed_flux = computed_flux[1:]

            if after_time_queries is not None:
                # Compute the flux values after the model's last valid time.
                after_time_mask = times > max_valid_time
                extrapolated_values = self._time_extrap_after.extrapolate_time(
                    max_valid_time,
                    np.array([computed_flux[-1]]),
                    after_time_queries,
                )
                new_computed_flux[after_time_mask] = extrapolated_values[:, 0]
                in_bounds_mask[after_time_mask] = False

                # Drop the last entry (which was added for extrapolation).
                computed_flux = computed_flux[:-1]

            # Fill in the valid flux values.
            new_computed_flux[in_bounds_mask] = computed_flux
            computed_flux = new_computed_flux

        return computed_flux

    def _evaluate_bandfluxes_single(self, passband_group, times, filters, state, rng_info=None) -> np.ndarray:
        """Get the band fluxes for a given PassbandGroup and a single, given graph state.

        Note
        ----
        This function does not compute SEDs and integrate them through the passbands, but
        rather uses band fluxes directly.

        Parameters
        ----------
        passband_group : PassbandGroup
            The passband group to use.
        times : numpy.ndarray
            A length T array of observer frame timestamps in MJD.
        filters : numpy.ndarray
            A length T array of filter names.
        state : GraphState
            An object mapping graph parameters to their values.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.

        Returns
        -------
        bandflux : numpy.ndarray
            A length T array of band fluxes for this sample.
        """
        params = self.get_local_params(state)

        # Compute the bandflux for each filter.
        bandfluxes = np.zeros(len(times))
        for filter_name in np.unique(filters):
            filter_mask = filters == filter_name
            bandfluxes[filter_mask] = self.compute_bandflux_with_extrapolation(
                times[filter_mask],
                filter_name,
                state,
                rng_info=rng_info,
            )

        # Apply all effects. Note that BandfluxModel does not apply redshift, so all effects
        # are applied in observer frame.
        for effect in self.band_pass_effects:
            bandfluxes = effect.apply_bandflux(
                bandfluxes,
                times=times,
                filters=filters,
                rng_info=rng_info,
                **params,  # Provide all the node's parameters to the effect.
            )
        return bandfluxes



import numpy as np
import pandas as pd
from lightcurvelynx.astro_utils.zeropoint import calculate_zp_from_maglim
from lightcurvelynx.consts import GAUSS_EFF_AREA2FWHM_SQ
from lightcurvelynx.obstable.obs_table import ObsTable
from lightcurvelynx.obstable.obs_table_params import (
    FiveSigmaDepthDeriver,
    FullParamDeriver,
    NoopParamDeriver,
)


def test_noop_param_deriver():
    """Use the NoopParamDeriver object to fill in missing ObsTable parameters."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    pdf = pd.DataFrame(values)

    ops_data = ObsTable(pdf, seeing=0.5, pixel_scale=0.2, sky_bg_adu=100.0, gain=2.0)
    assert len(ops_data) == 5

    # The table contains only the information provided.
    given_keys = ["time", "ra", "dec", "filter", "seeing", "pixel_scale", "sky_bg_adu", "gain"]
    not_given = ["zp_per_band", "exptime", "nexposure", "zp", "psf_footprint", "fwhm_px", "sky_bg_electrons"]
    assert np.all([key in ops_data for key in given_keys])
    assert np.all([key not in ops_data for key in not_given])

    # We can derive additional parameters.
    deriver = NoopParamDeriver()
    deriver.derive_parameters(ops_data)

    # Original keys
    assert np.all([key in ops_data for key in given_keys])

    # No new keys should be added.
    assert np.all([key not in ops_data for key in not_given])


def test_full_param_deriver():
    """Use the FullParamDeriver object to fill in missing ObsTable parameters."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
    }
    pdf = pd.DataFrame(values)

    zp_per_band = {"g": 26.0, "r": 27.0, "i": 28.0}
    ops_data = ObsTable(pdf, zp_per_band=zp_per_band, seeing=0.5, pixel_scale=0.2, sky_bg_adu=100.0, gain=2.0)
    assert len(ops_data) == 5

    # The table contains only the information provided.
    given_keys = ["time", "ra", "dec", "filter", "zp_per_band", "seeing", "pixel_scale", "sky_bg_adu", "gain"]
    assert np.all([key in ops_data for key in given_keys])
    assert np.all([key not in ops_data for key in ["exptime", "nexposure", "zp", "psf_footprint", "fwhm_px"]])

    # We can derive additional parameters.
    deriver = FullParamDeriver()
    deriver.derive_parameters(ops_data)

    # Original keys
    assert np.all([key in ops_data for key in given_keys])

    # Derived keys (one step of derivation)
    assert "zp" in ops_data
    assert np.allclose(ops_data["zp"], np.array([27.0, 26.0, 27.0, 28.0, 26.0]))

    assert "sky_bg_electrons" in ops_data
    assert np.allclose(ops_data["sky_bg_electrons"], np.array([200.0, 200.0, 200.0, 200.0, 200.0]))

    assert "fwhm_px" in ops_data
    assert np.allclose(ops_data["fwhm_px"], np.array([2.5] * 5))

    # Derived keys (two steps of derivation)
    assert "psf_footprint" in ops_data
    assert np.allclose(ops_data["psf_footprint"], np.array([GAUSS_EFF_AREA2FWHM_SQ * (2.5) ** 2] * 5))


def test_full_param_deriver_zp():
    """Use the FullParamDeriver object to compute a non-trivial zero point."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
        "maglim": np.array([22.0] * 5),
    }
    pdf = pd.DataFrame(values)

    ops_data = ObsTable(
        pdf,
        maglim=22.0,
        sky_bg_electrons=1000.0,
        fwhm_px=3.0,
        read_noise=5.0,
        dark_current=0.01,
        nexposure=1,
        exptime=30.0,
    )
    assert len(ops_data) == 5

    # The table contains only the information provided.
    given_cols = ["time", "ra", "dec", "filter", "maglim"]
    assert np.all([col in ops_data for col in given_cols])
    given_params = ["sky_bg_electrons", "fwhm_px", "read_noise", "dark_current", "exptime", "nexposure"]
    assert np.all([param in ops_data for param in given_params])
    assert np.all([key not in ops_data for key in ["zp", "psf_footprint", "seeing"]])

    # We can derive additional parameters.
    deriver = FullParamDeriver()
    deriver.derive_parameters(ops_data)

    # Derived keys (one step of derivation)
    assert "zp" in ops_data
    expected_zp = calculate_zp_from_maglim(
        maglim=22.0,
        sky_bg_electrons=1000.0,
        fwhm_px=3.0,
        read_noise=5.0,
        dark_current=0.01,
        exptime=30.0,
        nexposure=1,
    )
    assert np.allclose(ops_data["zp"], expected_zp)


def test_five_sigma_depth_deriver():
    """Use the FiveSigmaDepthDeriver object to compute the five-sigma depth."""
    values = {
        "time": np.array([0.0, 1.0, 2.0, 3.0, 4.0]),
        "ra": np.array([15.0, 30.0, 15.0, 0.0, 60.0]),
        "dec": np.array([-10.0, -5.0, 0.0, 5.0, 10.0]),
        "filter": np.array(["r", "g", "r", "i", "g"]),
        "five_sigma_depth": 20.0 + np.arange(5),
    }
    pdf = pd.DataFrame(values)

    bandflux_ref = {"g": 3630e6, "r": 3635e6, "i": 3625e6}
    ops_data = ObsTable(pdf, bandflux_ref=bandflux_ref)
    assert len(ops_data) == 5

    # The table contains only the information provided.
    given_data = ["time", "ra", "dec", "filter", "five_sigma_depth", "bandflux_ref"]
    assert np.all([col in ops_data for col in given_data])
    assert np.all([key not in ops_data for key in ["zp", "bandflux_error", "seeing"]])

    # We can derive additional parameters.
    deriver = FiveSigmaDepthDeriver()
    deriver.derive_parameters(ops_data)

    # Derived keys (one step of derivation)
    assert "bandflux_error" in ops_data
    bandflux_ref_arr = np.array([bandflux_ref[filt] for filt in ops_data["filter"]])
    expected_bandflux_error = (
        bandflux_ref_arr * np.power(10.0, -0.4 * ops_data["five_sigma_depth"].to_numpy()) / 5.0
    )
    assert np.allclose(ops_data["bandflux_error"], expected_bandflux_error)



import numpy as np

from lightcurvelynx.effects.effect_model import EffectModel


class WhiteNoise(EffectModel):
    """A white noise model.

    Attributes
    ----------
    white_noise_sigma : parameter
        The scale of the noise.
    """

    def __init__(self, white_noise_sigma, **kwargs):
        super().__init__(**kwargs)
        self.add_effect_parameter("white_noise_sigma", white_noise_sigma)

    def apply(
        self,
        flux_density,
        times=None,
        wavelengths=None,
        white_noise_sigma=None,
        rng_info=None,
        **kwargs,
    ):
        """Apply the effect to observations (flux_density values).

        Parameters
        ----------
        flux_density : numpy.ndarray
            A length T X N matrix of flux density values (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD). Not used for this effect.
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms). Not used for this effect.
        white_noise_sigma : float, optional
            The scale of the noise. Raises an error if None is provided.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of flux densities after the effect is applied (in nJy).
        """
        if white_noise_sigma is None:
            raise ValueError("white_noise_sigma must be provided")

        if rng_info is None:
            rng_info = np.random.default_rng()
        return rng_info.normal(loc=flux_density, scale=white_noise_sigma)

    def apply_bandflux(
        self,
        bandfluxes,
        *,
        times=None,
        filters=None,
        white_noise_sigma=None,
        rng_info=None,
        **kwargs,
    ):
        """Apply the effect to band fluxes.

        Parameters
        ----------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD).
        filters : numpy.ndarray, optional
            A length N array of filters. If not provided, the effect is applied to all
            band fluxes.
        white_noise_sigma : float, optional
            The scale of the noise. Raises an error if None is provided.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes after the effect is applied (in nJy).
        """
        if white_noise_sigma is None:
            raise ValueError("white_noise_sigma must be provided")

        if rng_info is None:
            rng_info = np.random.default_rng()
        return rng_info.normal(loc=bandfluxes, scale=white_noise_sigma)



from __future__ import annotations  # "type1 | type2" syntax in Python <3.9

import numpy as np
import numpy.typing as npt

from lightcurvelynx.astro_utils.mag_flux import mag2flux


def sky_bg_adu_to_electrons(sky_bg_adu, gain):
    """Convert sky background from ADU/pixel to electrons/pixel.

    Parameters
    ----------
    sky_bg_adu : float or ndarray of float
        Sky background in ADU/pixel.
    gain : float or ndarray of float
        The CCD gain (in e-/ADU).

    Returns
    -------
    float or ndarray of float
        Sky background in electrons/pixel.
    """
    return sky_bg_adu * gain


def magnitude_electron_zeropoint(
    *,
    filter: npt.ArrayLike,
    airmass: npt.ArrayLike,
    exptime: npt.ArrayLike,
    instr_zp_mag: dict[str, float] | float | npt.ArrayLike,
    ext_coeff: dict[str, float] | float | npt.ArrayLike,
) -> npt.ArrayLike:
    """Photometric zeropoint (magnitude that produces 1 electron) for
    LSST bandpasses (v1.9), using a standard atmosphere scaled
    for different airmasses and scaled for exposure times.

    Parameters
    ----------
    filter : ndarray of str
        The filter for which to return the photometric zeropoint.
    airmass : ndarray of float
        The airmass at which to return the photometric zeropoint.
    exptime : ndarray of float
        The exposure time for which to return the photometric zeropoint.
    instr_zp_mag : dict[str, float], float, or ndarray of float
        The instrumental zeropoint for each bandpass,
        i.e. AB-magnitude that produces 1 electron in a 1-second exposure.
        Keys are the bandpass names, values are the zeropoints.
    ext_coeff : dict[str, float], float, or ndarray of float
        Atmospheric extinction coefficient for each bandpass.
        Keys are the bandpass names, values are the coefficients.

    Returns
    -------
    ndarray of float
        AB mags that produces 1 electron.

    Notes
    -----
    Typically, zeropoints are defined as the magnitude of a source
    which would produce 1 count in a 1 second exposure -
    here we use *electron* counts, not ADU counts.

    References
    ----------
    Lynne Jones - https://community.lsst.org/t/release-of-v3-4-simulations/8548/12
    """
    # If we are given dictionaries mapping filter to value for either instr_zp_mag or ext_coeff,
    # convert them to the corresponding array of values for the input filter array.
    if isinstance(instr_zp_mag, dict):
        instr_zp_mag = np.vectorize(instr_zp_mag.get)(filter)
    if isinstance(ext_coeff, dict):
        ext_coeff = np.vectorize(ext_coeff.get)(filter)
    return instr_zp_mag + ext_coeff * (airmass - 1) + 2.5 * np.log10(exptime)


def flux_electron_zeropoint(
    *,
    instr_zp_mag: dict[str, float] | float | npt.ArrayLike,
    ext_coeff: dict[str, float] | float | npt.ArrayLike,
    filter: npt.ArrayLike,
    airmass: npt.ArrayLike,
    exptime: npt.ArrayLike,
) -> npt.ArrayLike:
    """Flux (nJy) producing 1 electron.

    Parameters
    ----------
    filter : npt.ArrayLike
        The filter for which to return the photometric zeropoint.
    airmass : npt.ArrayLike
        The airmass at which to return the photometric zeropoint.
    exptime : npt.ArrayLike
        The exposure time for which to return the photometric zeropoint.
    instr_zp_mag : dict[str, float], float, or ndarray of float
        The instrumental zeropoint for each bandpass in AB magnitudes,
        i.e. the magnitude that produces 1 electron in a 1-second exposure.
        Keys are the bandpass names, values are the zeropoints.
    ext_coeff : dict[str, float], float, or ndarray of float
        Atmospheric extinction coefficient for each bandpass.
        Keys are the bandpass names, values are the coefficients.

    Returns
    -------
    ndarray of float
        Flux (nJy) per electron.
    """
    mag_zp_electron = magnitude_electron_zeropoint(
        instr_zp_mag=instr_zp_mag,
        ext_coeff=ext_coeff,
        filter=filter,
        airmass=airmass,
        exptime=exptime,
    )
    return mag2flux(mag_zp_electron)


def calculate_zp_from_maglim(
    maglim=None,
    sky_bg_electrons=None,
    fwhm_px=None,
    read_noise=None,
    dark_current=None,
    exptime=None,
    nexposure=1,
):
    """Calculate zero points based on the 5-sigma mag limit.

    snr = flux/fluxerr
    fluxerr = sqrt(flux + sky*npix*gain + readnoise**2*nexposure*npix + darkcurrent*npix*exptime*nexposure)
    5 = flux/fluxerr
    25 = flux**2/(flux + sky*npix*Gain + readnoise**2*nexposure*npix + darkcurrent*npix*exptime*nexposure)
    flux**2 - 25*flux -25*( sky*npix*Gain
                            + readnoise**2*nexposure*npix
                            + darkcurrent*npix*exptime*nexposure)
                        = 0
    flux = 12.5 + 0.5*sqrt(625
                            + 100( sky*npix*Gain
                            + readnoise**2*nexposure*npix
                            + darkcurrent*npix*exptime*nexposure) )
    zp = 2.5*log10(flux) + maglim

    Parameters
    ----------
    maglim : float or ndarray
        Five-sigma magnitude limit.
    sky_bg_electrons : float or ndarray
        Sky background in electrons/pixel.
    fwhm_px : float or ndarray
        PSF in pixels.
    read_noise : float or ndarray
        Read noise (in e-/pixel).
    dark_current : float or ndarray
        Dark current (in e-/pixel/second).
    exptime : float or ndarray
        Exposure time (in seconds).
    nexposure : int or ndarray
        Number of exposure. Default is 1.

    Returns
    -------
    zp: float or ndarray
        Instrument zero point (that converts 1 e- to magnitude).
    """
    npix = 2.266 * fwhm_px**2  # = 4 * pi * sigma**2 = pi/2/ln2 * FWHM**2
    flux_at_5sigma_limit = 12.5 + 2.5 * np.sqrt(
        25.0
        + 4.0
        * (
            sky_bg_electrons * npix
            + read_noise**2 * nexposure * npix
            + dark_current * npix * exptime * nexposure
        )
    )
    zp = 2.5 * np.log10(flux_at_5sigma_limit) + maglim

    return zp



"""Samplers used for generating (RA, dec) coordinates."""

import logging
from pathlib import Path

import numpy as np
from astropy.coordinates import Angle, SkyCoord
from cdshealpix.nested import healpix_to_skycoord
from citation_compass import CiteClass
from mocpy import MOC

from lightcurvelynx.math_nodes.given_sampler import TableSampler
from lightcurvelynx.math_nodes.np_random import NumpyRandomFunc
from lightcurvelynx.obstable.obs_table import ObsTable


class UniformRADEC(NumpyRandomFunc):
    """A FunctionNode that uniformly samples (RA, dec) over a sphere,

    Attributes
    ----------
    use_degrees : bool
        The default return unit. If True returns samples in degrees.
        Otherwise, if False, returns samples in radians.
    """

    def __init__(self, outputs=None, seed=None, use_degrees=True, **kwargs):
        self.use_degrees = use_degrees

        # Override key arguments. We create a uniform sampler function, but
        # won't need it because the subclass overloads compute().
        func_name = "uniform"
        outputs = ["ra", "dec"]
        super().__init__(func_name, outputs=outputs, seed=seed, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        rng = rng_info if rng_info is not None else self._rng

        # Generate the random (RA, dec) lists.
        ra = rng.uniform(0.0, 2.0 * np.pi, size=graph_state.num_samples)
        dec = np.arcsin(rng.uniform(-1.0, 1.0, size=graph_state.num_samples))
        if self.use_degrees:
            ra = np.degrees(ra)
            dec = np.degrees(dec)

        # If we are generating a single sample, return floats.
        if graph_state.num_samples == 1:
            ra = ra[0]
            dec = dec[0]

        # Set the outputs and return the results. This takes the place of
        # function node's _save_results() function because we know the outputs.
        graph_state.set(self.node_string, "ra", ra)
        graph_state.set(self.node_string, "dec", dec)
        return [ra, dec]


class ObsTableRADECSampler(TableSampler):
    """A FunctionNode that samples RA and dec (and any extra columns) from around
    given data. This node is used for both sampling from a list of pointings (where
    the radius is the field of view) or sampling from a list of true objects.

    Parameters
    ----------
    data : ObsTable, Pandas DataFrame, NestedFrame, or dict
        The data to use for sampling. Must contain 'ra' and 'dec' columns.
    extra_cols : list of str, optional
        A list of extra column names to include in the sampling.
        Default: None
    radius : float, optional
        The sampling radius around the given points. If the points represent the
        center of a pointing, this is the radius of the field of view in degrees.
        If the points represent exact true objects, this can be a noise factor.
        Use 0.0 to return the exact given points.
        If None and data is an ObsTable, uses the value from the ObsTable.
        Default: None
    in_order : bool
        Return the given data in order of the rows (True). If False, performs
        random sampling with replacement.
        Default: False
    **kwargs : dict, optional
        Additional keyword arguments to pass to the parent class constructor.
    """

    def __init__(self, data, *, extra_cols=None, radius=None, in_order=False, **kwargs):
        if isinstance(data, ObsTable):
            if radius is None:
                radius = data.radius
            else:
                logging.getLogger(__name__).info(
                    f"Using provided radius {radius} instead of ObsTable radius {data.radius}."
                )
        if radius is None or radius < 0.0:
            raise ValueError(f"Invalid radius: {radius}")
        self.radius = radius

        # Start with RA, dec, and (optionally) time.
        data_dict = {
            "ra": data["ra"],
            "dec": data["dec"],
        }
        if "time" in data:
            data_dict["time"] = data["time"]

        # Add any extra columns (without duplicates).
        if extra_cols is not None:
            for col in extra_cols:
                if col not in data_dict:
                    data_dict[col] = data[col]

        super().__init__(data_dict, in_order=in_order, **kwargs)

    @classmethod
    def from_hats(cls, path, *, radius=None, extra_cols=None, in_order=False, **kwargs):
        """Create a GivenRADECSampler from the observations in a HATS Catalog.

        Note
        ----
        If you have an existing Dask client, it may be used.
        See LSDB documentation for details: https://docs.lsdb.io/en/latest/

        Parameters
        ----------
        path : str or Path
            The base path of the HATS data directory.
        radius : float, optional
            The sampling radius (noise factor) in degrees. Use 0.0 to return
            the exact pointings.
            Default: 0.0
        extra_cols : list of str, optional
            A list of extra column names to include in the sampling.
            Default: None
        in_order : bool
            Return the given data in order of the rows (True). If False, performs
            random sampling with replacement.
            Default: False
        **kwargs : dict, optional
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        GivenRADECSampler
            The created GivenRADECSampler object.
        """
        # See if the (optional) LSDB package is installed.
        try:
            from lsdb import read_hats
        except ImportError as err:  # pragma: no cover
            raise ImportError(
                "The lsdb package is required to read HATS catalogs. "
                "Please install it via 'pip install lsdb' or 'conda install conda-forge::lsdb'."
            ) from err

        # Compute the full list of columns to load from HATS.
        cols_to_load = ["ra", "dec"]
        if extra_cols is not None:
            cols_to_load.extend(extra_cols)
        columns = list(set(cols_to_load))  # Remove any duplicates.

        data = read_hats(path, columns=columns).compute()
        return cls(data, extra_cols=extra_cols, in_order=in_order, radius=radius, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        results : any
            The result of the computation. This return value is provided so that testing
            functions can easily access the results.
        """
        # Sample the center RA, dec, and times without the radius. Results is a vector
        # of arrays, one per output column. The first two are guaranteed to be RA and dec.
        results = super().compute(graph_state, rng_info=rng_info, **kwargs)

        if self.radius > 0.0:
            rng = rng_info if rng_info is not None else self._rng
            center = SkyCoord(ra=results[0], dec=results[1], unit="deg")

            # Add an offset from the center of the pointing defined by an offset angle (phi)
            # and offset radius (theta).  Both are defined in radians.
            offset_dir = rng.uniform(0.0, 2.0 * np.pi, size=graph_state.num_samples)
            cos_radius = np.cos(np.radians(self.radius))
            offset_amt = np.arccos(rng.uniform(cos_radius, 1.0, size=graph_state.num_samples))
            new_coords = center.directional_offset_by(
                position_angle=Angle(offset_dir, "radian"),
                separation=Angle(offset_amt, "radian"),
            )

            # Replace the results' RA and dec with the new pointing (in degrees).
            results[0] = new_coords.ra.deg
            results[1] = new_coords.dec.deg

            # Resave the results to transfer them to the graph state.
            self._save_results(results, graph_state)

        return results


class ObsTableUniformRADECSampler(NumpyRandomFunc):
    """A FunctionNode that samples RA and dec uniformly from the area covered
    by an ObsTable.  RA and dec are returned in degrees.

    Note
    ----
    This uses rejection sampling where it randomly guesses an (RA, dec) then checks if that
    point falls within the survey. If not, it repeats the process until it finds a valid point
    or reaches `max_iterations` iterations (then returns the last sample). This sampling method
    can be quite slow or even generate out-of-survey samples if the coverage is small.

    Attributes
    ----------
    data : ObsTable
        The ObsTable object to use for sampling.
    radius : float
        The radius of the field of view of the observations in degrees.
    max_iterations : int
        The maximum number of iterations to perform. Default: 1000

    Parameters
    ----------
    data : ObsTable
        The ObsTable object to use for sampling.
    radius : float, optional
        The search radius around the center of the pointing. If None, uses the
        value from the ObsTable.
    outputs : list of str, optional
        The list of output names. Default: ["ra", "dec"]
    seed : int, optional
        The random seed to use for the internal random number generator. Default: None
    max_iterations : int, optional
        The maximum number of iterations to perform. Default: 1000
    **kwargs : dict, optional
        Additional keyword arguments to pass to the parent class constructor.
    """

    def __init__(self, data, *, radius=None, outputs=None, seed=None, max_iterations=1000, **kwargs):
        if isinstance(data, ObsTable):
            if radius is None:
                radius = data.radius
            else:
                logging.getLogger(__name__).info(
                    f"Using provided radius {radius} instead of ObsTable radius {data.radius}."
                )
        if radius is None or radius < 0.0:
            raise ValueError(f"Invalid radius: {radius}")
        self.radius = radius

        if len(data) == 0:
            raise ValueError("ObsTable data cannot be empty.")
        self.data = data

        if max_iterations <= 0:
            raise ValueError("Invalid max_iterations: {max_iterations}")
        self.max_iterations = max_iterations

        # Override key arguments. We create a uniform sampler function, but
        # won't need it because the subclass overloads compute().
        func_name = "uniform"
        outputs = ["ra", "dec"]
        super().__init__(func_name, outputs=outputs, seed=seed, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        (ra, dec) : tuple of floats or np.ndarray
            If a single sample is generated, returns a tuple of floats. Otherwise,
            returns a tuple of np.ndarrays.
        """
        rng = rng_info if rng_info is not None else self._rng

        ra = np.zeros(graph_state.num_samples)
        dec = np.zeros(graph_state.num_samples)
        mask = np.full(graph_state.num_samples, False)
        num_missing = graph_state.num_samples

        # Rejection sampling to ensure the samples are within the ObsTable coverage.
        # This can take many iterations if the coverage is small.
        iter_num = 1
        while num_missing > 0 and iter_num <= self.max_iterations:
            # Generate new samples for the missing ones.
            new_ra = np.degrees(rng.uniform(0.0, 2.0 * np.pi, size=num_missing))
            new_dec = np.degrees(np.arcsin(rng.uniform(-1.0, 1.0, size=num_missing)))
            ra[~mask] = new_ra
            dec[~mask] = new_dec

            # Check if the new samples are within the ObsTable coverage. We don't
            # need to recheck the ones that were already valid.
            mask[~mask] = self.data.is_observed(new_ra, new_dec, radius=self.radius)
            num_missing = np.sum(~mask)
            iter_num += 1

        # If we are generating a single sample, return floats.
        if graph_state.num_samples == 1:
            ra = ra[0]
            dec = dec[0]

        # Set the outputs and return the results. This takes the place of
        # function node's _save_results() function because we know the outputs.
        graph_state.set(self.node_string, "ra", ra)
        graph_state.set(self.node_string, "dec", dec)
        return (ra, dec)


class ApproximateMOCSampler(NumpyRandomFunc, CiteClass):
    """A FunctionNode that samples RA and dec (approximately) from the coverage of
    a MOCPy Multi-Order Coverage Map object.

    References
    ----------
    * MOCPY: https://github.com/cds-astro/mocpy/
    * CDS Healpix: https://github.com/cds-astro/cds-healpix-python
    * MOC: Pierre Fernique, Thomas Boch, Tom Donaldson, Daniel Durand , Wil O'Mullane, Martin Reinecke,
    and Mark Taylor. MOC - HEALPix Multi-Order Coverage map Version 1.0. IVOA Recommendation 02 June 2014,
    pages 602, Jun 2014. doi:10.5479/ADS/bib/2014ivoa.spec.0602F.

    Attributes
    ----------
    healpix_list : list of int
        The list of healpix pixel IDs that cover the MOC at the given depth.
    depth : int
        The healpix depth to use as an approximation. Must be [2, 29].
    """

    def __init__(self, moc, *, outputs=None, seed=None, depth=12, **kwargs):
        if depth < 2 or depth > 29:
            raise ValueError("Depth must be [2, 29]. Received {depth}")
        self.depth = depth
        self.healpix_list = moc.to_order(depth).flatten()

        # Override key arguments. We create a uniform sampler function, but
        # won't need it because the subclass overloads compute().
        func_name = "uniform"
        outputs = ["ra", "dec"]
        super().__init__(func_name, outputs=outputs, seed=seed, **kwargs)

    @classmethod
    def from_file(cls, filename, format="fits", **kwargs):
        """Create an ApproximateMOCSampler from a MOC file.

        This file can be created from a mocpy.MOC object using its save() function.

        Parameters
        ----------
        filename : str or Path
            The path to the MOC file. Supported formats include FITS, JSON, and ASCII.
        format : str, optional
            The format of the MOC file. Supported formats include 'fits', 'json', and
            'ascii'. Default is 'fits'.
        **kwargs : dict, optional
            Additional keyword arguments to pass to the constructor.

        Returns
        -------
        ApproximateMOCSampler
            The created ApproximateMOCSampler object.
        """
        filename = Path(filename)
        if not filename.is_file():
            raise FileNotFoundError(f"MOC file not found: {filename}")

        moc = MOC.load(filename, format=format)
        return cls(moc, **kwargs)

    def compute(self, graph_state, rng_info=None, **kwargs):
        """Return the given values.

        Parameters
        ----------
        graph_state : GraphState
            An object mapping graph parameters to their values. This object is modified
            in place as it is sampled.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : dict, optional
            Additional function arguments.

        Returns
        -------
        (ra, dec) : tuple of floats or np.ndarray
            If a single sample is generated, returns a tuple of floats. Otherwise,
            returns a tuple of np.ndarrays.
        """
        rng = rng_info if rng_info is not None else self._rng

        # Choose a starting pixel ID for each sample. Then randomly traverse
        # down the healpix tree by moving to one of the children pixels until
        # we reach level=29 (approximately 4.5 * 10^18 possible locations).
        pixel_ids = rng.choice(self.healpix_list, size=graph_state.num_samples).astype(np.uint64)
        start_pixel_ids29 = np.left_shift(pixel_ids, 2 * (29 - self.depth))
        offset_range = np.uint64(1) << np.uint64(2 * (29 - self.depth))
        pixel_ids29 = start_pixel_ids29 + rng.integers(
            offset_range, size=graph_state.num_samples, dtype=np.uint64
        )

        # Convert back the healpix centers to RA and dec.
        coords = healpix_to_skycoord(pixel_ids29, depth=29)
        ra = coords.ra.deg
        dec = coords.dec.deg

        # If we are generating a single sample, return floats.
        if graph_state.num_samples == 1:
            ra = ra[0]
            dec = dec[0]

        # Set the outputs and return the results. This takes the place of
        # function node's _save_results() function because we know the outputs.
        graph_state.set(self.node_string, "ra", ra)
        graph_state.set(self.node_string, "dec", dec)
        return (ra, dec)



"""The base EffectModel class used for all effects."""


class EffectModel:
    """A physical or systematic effect to apply to an observation.

    Effects are not ParameterizedNodes but can have arguments that are
    ParameterizedNodes. These arguments are stored as parameters in the
    model node (BasePhysicalModel) when add_effect() is called. This allows
    the effect to easily use existing parameters from the model node as
    well as new parameters specific to the effect. The model node samples
    these parameters and passes them in the call to apply().

    The EffectModel class is designed to be subclassed, and the
    subclasses should implement the apply() method to define the
    specific effect being applied.

    Attributes
    ----------
    rest_frame : bool
        Whether the effect is applied in the rest frame of the observation (True)
        or in the observed frame (False).
    effect_name : string
        The name of the effect for logging and accessing parameters. If no name is
        provided, the class name is used.
    parameters : dict
        A dictionary of parameters for the effect. Maps the parameter names to
        their setters.
    """

    def __init__(self, *, rest_frame=True, effect_name=None, **kwargs):
        self.rest_frame = rest_frame

        # Save the name of this effect.
        if effect_name is None:
            effect_name = self.__class__.__name__
            effect_name = effect_name.replace(".", "_")
        self.effect_name = effect_name

        # Automatically include all keyword arguments as settable parameters.
        self.parameters = {}
        for key, value in kwargs.items():
            self.add_effect_parameter(key, value)

    def __str__(self):
        return self.effect_name

    def __repr__(self):
        # We only include the parameter names because the values may be dynamically
        # sampled during the simulation.
        params = ",".join(self.parameters.keys())
        return f"{self.effect_name}({params})"

    def add_effect_parameter(self, name, setter):
        """Add a parameter to the effect.

        Note
        ----
        These parameters are automatically added to the corresponding model
        nodes so they are sampled and recorded with the model's other parameters.

        Parameters
        ----------
        name : str
            The name of the parameter.
        setter : function
            A function that sets the parameter value.
        """
        self.parameters[name] = setter

    def apply(self, flux_density, times=None, wavelengths=None, rng_info=None, **kwargs):
        """Apply the effect to observations (flux_density values).

        Parameters
        ----------
        flux_density : numpy.ndarray
            A length T X N matrix of flux density values (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD).
        wavelengths : numpy.ndarray, optional
            A length N array of wavelengths (in angstroms).
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        flux_density : numpy.ndarray
            A length T x N matrix of flux densities after the effect is applied (in nJy).
        """
        raise NotImplementedError()

    def apply_bandflux(self, bandfluxes, *, times=None, filters=None, rng_info=None, **kwargs):
        """Apply the effect to band fluxes.

        Parameters
        ----------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes (in nJy).
        times : numpy.ndarray, optional
            A length T array of times (in MJD).
        filters : numpy.ndarray, optional
            A length N array of filters. If not provided, the effect is applied to all
            band fluxes.
        rng_info : numpy.random._generator.Generator, optional
            A given numpy random number generator to use for this computation. If not
            provided, the function uses the node's random number generator.
        **kwargs : `dict`, optional
           Any additional keyword arguments, including any additional
           parameters needed to apply the effect.

        Returns
        -------
        bandfluxes : numpy.ndarray
            A length T array of band fluxes after the effect is applied (in nJy).
        """
        raise NotImplementedError()



"""Functions for extrapolating flux past the end of a model's range of valid
phases or wavelengths using flux = f(time, wavelengths).
"""

import abc

import numpy as np


class FluxExtrapolationModel(abc.ABC):
    """The base class for the flux extrapolation methods."""

    @abc.abstractmethod
    def __init__(self):
        pass

    @abc.abstractmethod
    def _extrapolate(self, last_value, last_flux, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_flux) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_value : float
            The last valid value along the extrapolation axis a which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_flux : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """
        raise NotImplementedError("Subclasses must implement this method.")

    def extrapolate_time(self, last_time, last_flux, query_times):
        """Extrapolate along the time axis.

        Parameters
        ----------
        last_time : float
            The last valid time (in days) at which the flux was predicted.
        last_flux : numpy.ndarray
            A length W array of the last valid flux values at each wavelength
            at the last valid time (in nJy).
        query_times : numpy.ndarray
            A length T array of the query times (in days) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T x W matrix of extrapolated values.
        """
        return self._extrapolate(last_time, last_flux, query_times).T

    def extrapolate_wavelength(self, last_wave, last_flux, query_waves):
        """Extrapolate along the wavelength axis.

        Parameters
        ----------
        last_wave : float
            The last valid wavelength (in AA) at which the flux was predicted.
        last_flux : numpy.ndarray
            A length T array of the last valid flux values at each time
            at the last valid wavelength (in nJy).
        query_waves : numpy.ndarray
            A length W array of the query wavelengths (in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A T x W matrix of extrapolated values.
        """
        # We transpose the result to turn the W x T matrix into a T x W matrix.
        return self._extrapolate(last_wave, last_flux, query_waves)


class ZeroPadding(FluxExtrapolationModel):
    """Extrapolate by zero padding the results."""

    def __init__(self):
        super().__init__()

    def _extrapolate(self, last_value, last_flux, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_flux) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_value : float
            The last valid value along the extrapolation axis a which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_flux : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """
        N_len = len(last_flux)
        M_len = len(query_values)
        return np.zeros((N_len, M_len))


class ConstantPadding(FluxExtrapolationModel):
    """Extrapolate using a constant value in nJy.

    Attributes
    ----------
    value : float
        The value (in nJy) to use for the extrapolation.
    """

    def __init__(self, value=0.0):
        super().__init__()

        if value < 0:
            raise ValueError("Extrapolation value must be positive.")
        self.value = value

    def _extrapolate(self, last_value, last_flux, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_flux) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_value : float
            The last valid value along the extrapolation axis a which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_flux : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """
        N_len = len(last_flux)
        M_len = len(query_values)
        return np.full((N_len, M_len), self.value)


class LastValue(FluxExtrapolationModel):
    """Extrapolate using the last valid value along the desired axis."""

    def __init__(self):
        super().__init__()

    def _extrapolate(self, last_value, last_flux, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_flux) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_value : float
            The last valid value along the extrapolation axis a which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_flux : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """
        return np.tile(last_flux[:, np.newaxis], (1, len(query_values)))


class LinearDecay(FluxExtrapolationModel):
    """Linear decay of the flux using the last valid point(s) down to zero.

    Attributes
    ----------
    decay_width : float
        The width of the decay region in Angstroms. The flux is
        linearly decreased to zero over this range.
    """

    def __init__(self, decay_width=100.0):
        super().__init__()

        if decay_width <= 0:
            raise ValueError("decay_width must be positive.")
        self.decay_width = decay_width

    def _extrapolate(self, last_value, last_flux, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_flux) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_value : float
            The last valid value along the extrapolation axis a which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_flux : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """
        last_flux = np.asarray(last_flux)
        query_values = np.asarray(query_values)
        dist = np.abs(query_values - last_value)

        multiplier = np.clip(1.0 - (dist / self.decay_width), 0.0, 1.0)
        flux = last_flux[:, np.newaxis] * multiplier[np.newaxis, :]
        return flux


class ExponentialDecay(FluxExtrapolationModel):
    """Exponential decay of the flux using the last valid point(s) down to zero.

    f(t, λ) = f(t, λ_last) * exp(- rate * \\|λ - λ_last\\|)

    Attributes
    ----------
    rate : float
        The decay rate in the exponential function.
    """

    def __init__(self, rate):
        super().__init__()

        if rate < 0:
            raise ValueError("Decay rate must be zero or positive.")
        self.rate = rate

    def _extrapolate(self, last_value, last_flux, query_values):
        """Evaluate the extrapolation given the last valid points(s) and a list of new
        query points.

        Note
        ----
        This function does not care which axis is being extrapolated. The returned values are
        always len(query_values) x len(last_flux) and may need to be transposed by the calling
        function.

        Parameters
        ----------
        last_value : float
            The last valid value along the extrapolation axis a which the flux was predicted
            (e.g., wavelength in AA or time in days).
        last_flux : numpy.ndarray
            A length N array of the flux values at the last valid time or wavelength (in nJy).
        query_values : numpy.ndarray
            A length M array of values along the extrapolation axis (times in days or wavelengths
            in AA) at which to extrapolate.

        Returns
        -------
        flux : numpy.ndarray
            A N x M matrix of extrapolated values. Where M is the number of query points and
            N is the number of flux values at the last valid point.
        """
        last_flux = np.asarray(last_flux)
        query_values = np.asarray(query_values)
        dist = np.abs(query_values - last_value)

        multiplier = np.exp(-self.rate * dist)
        flux = last_flux[:, np.newaxis] * multiplier[np.newaxis, :]
        return flux
